#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import ast
import importlib.machinery
import os
import sys
import py_compile
import textwrap
import zipfile

import pytest

from PyInstaller.lib.modulegraph import modulegraph
from PyInstaller.utils.tests import xfail


def _import_and_get_node(tmp_path, module_name, path=None):
    script = tmp_path / 'script.py'
    script.write_text(f"import {module_name}", encoding='utf-8')
    if path is None:
        path = [str(tmp_path)]
    mg = modulegraph.ModuleGraph(path)
    mg.add_script(str(script))
    return mg.find_node(module_name)


def test_sourcefile(tmp_path):
    (tmp_path / 'source.py').write_text("###", encoding='utf-8')
    node = _import_and_get_node(tmp_path, 'source')
    assert isinstance(node, modulegraph.SourceModule)


def test_invalid_sourcefile(tmp_path):
    (tmp_path / 'invalid_source.py').write_text("invalid python-source code", encoding='utf-8')
    node = _import_and_get_node(tmp_path, 'invalid_source')
    assert isinstance(node, modulegraph.InvalidSourceModule)


def test_invalid_compiledfile(tmp_path):
    (tmp_path / 'invalid_compiled.pyc').write_text("invalid byte-code", encoding='utf-8')
    node = _import_and_get_node(tmp_path, 'invalid_compiled')
    assert isinstance(node, modulegraph.InvalidCompiledModule)


def test_builtin(tmp_path):
    node = _import_and_get_node(tmp_path, 'sys', path=sys.path)
    assert isinstance(node, modulegraph.BuiltinModule)


def test_extension(tmp_path):
    node = _import_and_get_node(tmp_path, '_ctypes', path=sys.path)
    assert isinstance(node, modulegraph.Extension)


def test_package(tmp_path):
    # Create package; stuff/__init__.py
    (tmp_path / 'stuff').mkdir()
    pysrc = tmp_path / 'stuff' / '__init__.py'
    pysrc.write_text("###", encoding='utf-8')
    # Analyze
    node = _import_and_get_node(tmp_path, 'stuff')
    assert node.__class__ is modulegraph.Package
    assert node.filename in (str(pysrc), str(pysrc) + 'c')
    assert node.packagepath == [str(pysrc.parent)]


#-- Extension modules


@pytest.mark.parametrize(
    "num, modname, expected_nodetype",
    (
        # package's __init__ module is an extension
        (1, "myextpkg", modulegraph.ExtensionPackage),
        # __init__.py beside the __init__ module being an extension
        (2, "myextpkg", modulegraph.ExtensionPackage),
        # Importing a module beside
        (3, "myextpkg.other", modulegraph.Extension),
        # sub-package's __init__ module is an extension
        (4, "myextpkg.subpkg", modulegraph.ExtensionPackage),
        # importing a module beside, but from a sub-package
        (5, "myextpkg.subpkg.other", modulegraph.Extension),
    )
)
def test_package_init_is_extension(tmp_path, num, modname, expected_nodetype):
    # Regression: Recursion too deep

    EXTENSION_SUFFIX = importlib.machinery.EXTENSION_SUFFIXES[0]

    def _write_module(*args):
        module_file = tmp_path.joinpath(*args)
        module_file.parent.mkdir(parents=True, exist_ok=True)
        module_file.write_text('###', encoding='utf-8')
        return module_file

    def create_package_files(test_case):
        m = _write_module('myextpkg', '__init__' + EXTENSION_SUFFIX)
        if test_case == 1:
            return m
        _write_module('myextpkg', '__init__.py')
        if test_case == 2:
            return m  # return extension module anyway
        m = _write_module('myextpkg', 'other.py')
        m = _write_module('myextpkg', 'other' + EXTENSION_SUFFIX)
        if test_case == 3:
            return m
        m = _write_module('myextpkg', 'subpkg', '__init__.py')
        m = _write_module('myextpkg', 'subpkg', '__init__' + EXTENSION_SUFFIX)
        if test_case == 4:
            return m
        m = _write_module('myextpkg', 'subpkg', 'other.py')
        m = _write_module('myextpkg', 'subpkg', 'other' + EXTENSION_SUFFIX)
        return m

    module_file = create_package_files(num)
    node = _import_and_get_node(tmp_path, modname)
    assert node.__class__ is expected_nodetype
    if expected_nodetype is modulegraph.ExtensionPackage:
        assert node.packagepath == [str(module_file.parent)]
    else:
        assert node.packagepath is None  # not a package
    assert node.filename == str(module_file)


#-- Basic tests - these seem to be missing in the original modulegraph test-suite


def test_relative_import_missing(tmp_path):
    libdir = tmp_path / 'lib'

    pkg = libdir / 'pkg'
    (pkg / 'x' / 'y').mkdir(parents=True)  # Create the whole package directory tree
    (pkg / '__init__.py').write_text("#", encoding='utf-8')
    (pkg / 'x' / '__init__.py').write_text("#", encoding='utf-8')
    (pkg / 'x' / 'y' / '__init__.py').write_text("#", encoding='utf-8')
    (pkg / 'x' / 'y' / 'z.py').write_text("from . import DoesNotExist", encoding='utf-8')

    script = tmp_path / 'script.py'
    script.write_text("import pkg.x.y.z", encoding='utf-8')

    mg = modulegraph.ModuleGraph([str(libdir)])
    mg.add_script(str(script))
    assert isinstance(mg.find_node('pkg.x.y.z'), modulegraph.SourceModule)
    assert isinstance(mg.find_node('pkg.x.y.DoesNotExist'), modulegraph.MissingModule)


#-- Tests with a single module in a zip-file


def test_zipped_module_source(tmp_path):
    pysrc = tmp_path / 'stuff.py'
    pysrc.write_text("###", encoding='utf-8')

    zipfilename = tmp_path / 'unstuff.zip'
    with zipfile.ZipFile(zipfilename, mode='w') as zfh:
        zfh.write(pysrc, 'stuff.py')

    node = _import_and_get_node(tmp_path, 'stuff', path=[zipfilename])
    assert node.__class__ is modulegraph.SourceModule
    assert node.filename.startswith(os.path.join(zipfilename, 'stuff.py'))


def test_zipped_module_source_and_compiled(tmp_path):
    pysrc = tmp_path / 'stuff.py'
    pysrc.write_text("###", encoding='utf-8')

    pyc = pysrc.with_suffix('.pyc')
    py_compile.compile(pysrc, pyc)

    zipfilename = tmp_path / 'unstuff.zip'
    with zipfile.ZipFile(zipfilename, mode='w') as zfh:
        zfh.write(pysrc, 'stuff.py')
        zfh.write(pyc, 'stuff.pyc')

    node = _import_and_get_node(tmp_path, 'stuff', path=[zipfilename])
    # Do not care whether it is source or compiled, as long as it is neither invalid nor missing.
    assert node.__class__ in (modulegraph.SourceModule, modulegraph.CompiledModule)
    assert node.filename.startswith(os.path.join(zipfilename, 'stuff.py'))


#-- Tests with a package in a zip-file


def test_zipped_package_source(tmp_path):
    pkg = tmp_path / 'stuff'
    pkg.mkdir()

    pysrc = pkg / '__init__.py'
    pysrc.write_text('###', encoding='utf-8')

    zipfilename = tmp_path / 'stuff.zip'
    with zipfile.ZipFile(zipfilename, mode='w') as zfh:
        zfh.write(pkg, 'stuff')
        zfh.write(pysrc, 'stuff/__init__.py')

    node = _import_and_get_node(tmp_path, 'stuff', path=[zipfilename])
    assert node.__class__ is modulegraph.Package
    assert node.packagepath == [os.path.join(zipfilename, 'stuff')]


def test_zipped_package_source_and_compiled(tmp_path):
    pkg = tmp_path / 'stuff'
    pkg.mkdir()

    pysrc = pkg / '__init__.py'
    pysrc.write_text('###', encoding='utf-8')

    pyc = pysrc.with_suffix('.pyc')
    py_compile.compile(pysrc, pyc)

    zipfilename = tmp_path / 'stuff.zip'
    with zipfile.ZipFile(zipfilename, mode='w') as zfh:
        zfh.write(pkg, 'stuff')
        zfh.write(pysrc, 'stuff/__init__.py')
        zfh.write(pyc, 'stuff/__init__.pyc')

    node = _import_and_get_node(tmp_path, 'stuff', path=[zipfilename])
    assert node.__class__ is modulegraph.Package
    assert node.packagepath == [os.path.join(zipfilename, 'stuff')]


#-- Namespace packages


def test_nspackage_pep420(tmp_path):
    p1 = tmp_path / 'p1'
    (p1 / 'stuff').mkdir(parents=True)
    (p1 / 'stuff' / 'a.py').write_text("###", encoding='utf-8')

    p2 = tmp_path / 'p2'
    (p2 / 'stuff').mkdir(parents=True)
    (p2 / 'stuff' / 'b.py').write_text("###", encoding='utf-8')

    path = [str(p1), str(p2)]

    script = tmp_path / 'script.py'
    script.write_text("import stuff.a, stuff.b", encoding='utf-8')

    mg = modulegraph.ModuleGraph(path)
    mg.add_script(str(script))

    mg.report()

    assert isinstance(mg.find_node('stuff.a'), modulegraph.SourceModule)
    assert isinstance(mg.find_node('stuff.b'), modulegraph.SourceModule)

    node = mg.find_node('stuff')
    assert isinstance(node, modulegraph.NamespacePackage)
    assert node.packagepath == [os.path.join(p, 'stuff') for p in path]


# :todo: test_namespace_setuptools
# :todo: test_namespace_pkg_resources


@pytest.mark.darwin
@pytest.mark.linux
def test_symlinks(tmp_path):
    (tmp_path / 'p1').mkdir()
    p1_init = tmp_path / 'p1' / '__init__.py'
    p1_init.write_text("###", encoding='utf-8')

    (tmp_path / 'p2').mkdir()
    p2_init = tmp_path / 'p2' / '__init__.py'
    p2_init.write_text("###", encoding='utf-8')

    base_dir = tmp_path / 'base'
    (base_dir / 'p1').mkdir(parents=True)

    os.symlink(str(p1_init), str(base_dir / 'p1' / '__init__.py'))
    os.symlink(str(p2_init), str(base_dir / 'p1' / 'p2.py'))

    node = _import_and_get_node(base_dir, 'p1.p2')
    assert isinstance(node, modulegraph.SourceModule)


def test_import_order_1(tmp_path):
    # Ensure modulegraph processes modules in the same order as Python does.

    class MyModuleGraph(modulegraph.ModuleGraph):
        def _load_module(self, fqname, pathname, loader):
            if not record or record[-1] != fqname:
                record.append(fqname)  # record non-consecutive entries
            return super()._load_module(fqname, pathname, loader)

    record = []

    # (filename, content)
    ENTRIES = (
        ('a/', 'from . import c, d'),
        ('a/c', '#'),
        ('a/d/', 'from . import f, g, h'),
        ('a/d/f/', 'from . import j, k'),
        ('a/d/f/j', '#'),
        ('a/d/f/k', '#'),
        ('a/d/g/', 'from . import l, m'),
        ('a/d/g/l', '#'),
        ('a/d/g/m', '#'),
        ('a/d/h', '#'),
        ('b/', 'from . import e'),
        ('b/e/', 'from . import i'),
        ('b/e/i', '#'),
    )

    for filename, content in ENTRIES:
        if filename.endswith('/'):
            filename += '__init__'
        filename += '.py'
        module_fullpath = tmp_path / filename
        module_fullpath.parent.mkdir(parents=True, exist_ok=True)
        module_fullpath.write_text(content, encoding='utf-8')

    script = tmp_path / 'script.py'
    script.write_text("import a, b", encoding='utf-8')

    mg = MyModuleGraph([str(tmp_path)])
    mg.add_script(str(script))

    # This is the order Python imports these modules given that script.
    expected = [
        'a', 'a.c', 'a.d', 'a.d.f', 'a.d.f.j', 'a.d.f.k', 'a.d.g', 'a.d.g.l', 'a.d.g.m', 'a.d.h', 'b', 'b.e', 'b.e.i'
    ]
    assert record == expected


def test_import_order_2(tmp_path):
    # Ensure modulegraph processes modules in the same order as Python does.

    class MyModuleGraph(modulegraph.ModuleGraph):
        def _load_module(self, fqname, pathname, loader):
            if not record or record[-1] != fqname:
                record.append(fqname)  # record non-consecutive entries
            return super()._load_module(fqname, pathname, loader)

    record = []

    # (filename, content)
    ENTRIES = (
        ('a/', '#'),
        ('a/c/', '#'),
        ('a/c/g', '#'),
        ('a/c/h', 'from . import g'),
        ('a/d/', '#'),
        ('a/d/i', 'from ..c import h'),
        ('a/d/j/', 'from .. import i'),
        ('a/d/j/o', '#'),
        ('b/', 'from .e import k'),
        ('b/e/', 'import a.c.g'),
        ('b/e/k', 'from .. import f'),
        ('b/e/l', 'import a.d.j'),
        ('b/f/', '#'),
        ('b/f/m', '#'),
        ('b/f/n/', '#'),
        ('b/f/n/p', 'from ...e import l'),
    )

    for filename, content in ENTRIES:
        if filename.endswith('/'):
            filename += '__init__'
        filename += '.py'
        module_fullpath = tmp_path / filename
        module_fullpath.parent.mkdir(parents=True, exist_ok=True)
        module_fullpath.write_text(content, encoding='utf-8')

    script = tmp_path / 'script.py'
    script.write_text("import b.f.n.p", encoding='utf-8')

    mg = MyModuleGraph([str(tmp_path)])
    mg.add_script(str(script))

    # This is the order Python imports these modules given that script.
    expected = [
        'b', 'b.e', 'a', 'a.c', 'a.c.g', 'b.e.k', 'b.f', 'b.f.n', 'b.f.n.p', 'b.e.l', 'a.d', 'a.d.j', 'a.d.i', 'a.c.h'
    ]
    assert record == expected


#---- scan bytecode


def __scan_code(code, use_ast, monkeypatch):
    mg = modulegraph.ModuleGraph()
    # _process_imports would set _deferred_imports to None
    monkeypatch.setattr(mg, '_process_imports', lambda m: None)
    module = mg.createNode(modulegraph.Script, 'dummy.py')

    code = textwrap.dedent(code)
    if use_ast:
        co_ast = compile(code, 'dummy', 'exec', ast.PyCF_ONLY_AST)
        co = compile(co_ast, 'dummy', 'exec')
    else:
        co_ast = None
        co = compile(code, 'dummy', 'exec')
    mg._scan_code(module, co)
    return module


@pytest.mark.parametrize("use_ast", (True, False))
def test_scan_code__empty(monkeypatch, use_ast):
    code = "# empty code"
    module = __scan_code(code, use_ast, monkeypatch)
    assert len(module._deferred_imports) == 0
    assert len(module._global_attr_names) == 0


@pytest.mark.parametrize("use_ast", (True, False))
def test_scan_code__basic(monkeypatch, use_ast):
    code = """
    import os.path
    from sys import maxint, exitfunc, platform
    del exitfunc
    def testfunc():
        import shutil
    """
    module = __scan_code(code, use_ast, monkeypatch)
    assert len(module._deferred_imports) == 3
    assert ([di[1][0] for di in module._deferred_imports] == ['os.path', 'sys', 'shutil'])
    assert module.is_global_attr('maxint')
    assert module.is_global_attr('os')
    assert module.is_global_attr('platform')
    assert not module.is_global_attr('shutil')  # not imported at module level
    assert not module.is_global_attr('exitfunc')


#-- SWIG packages - pyinstaller specific tests


def _test_swig_import_simple_common(tmp_path):
    libdir = tmp_path / 'lib'

    osgeo = libdir / 'pyi_test_osgeo'
    osgeo.mkdir(parents=True)
    (osgeo / '__init__.py').write_text('#', encoding='utf-8')
    (osgeo / '_pyi_gdal.py').write_text("#", encoding='utf-8')
    (osgeo / 'pyi_gdal.py').write_text(
        "\n".join(["# automatically generated by SWIG", "import _pyi_gdal"]),
        encoding='utf-8',
    )

    script = tmp_path / 'script.py'
    script.write_text("from pyi_test_osgeo import pyi_gdal", encoding='utf-8')

    mg = modulegraph.ModuleGraph([str(libdir)])
    mg.add_script(str(script))

    assert isinstance(mg.find_node('pyi_test_osgeo'), modulegraph.Package)
    assert isinstance(mg.find_node('pyi_test_osgeo.pyi_gdal'), modulegraph.SourceModule)
    # The "C" module is frozen under its unqualified rather than qualified name.
    # See comment in modulegraph._safe_import_hook.
    # BUG: modulegraph contains a probable bug: Only the module's identifier is changed, not the module's graphident.
    # Thus the node is still found under its old name. The relevant code was brought from PyInstaller to upstream,
    # so this might be PyInstaller's fault. See test_swig_import_simple for what it should be.
    # This is a separate test-case, not marked as xfail, so we can spot whether the SWIG support works at all.
    assert isinstance(mg.find_node('pyi_test_osgeo._pyi_gdal'), modulegraph.SourceModule)
    # Due the the buggy implementation, the graphident is unchanged, but at least the identifier should have changed.
    assert mg.find_node('pyi_test_osgeo._pyi_gdal').identifier == '_pyi_gdal'
    # Due the the buggy implementation, this node does not exist.
    assert mg.find_node('_pyi_gdal') is None
    return mg  # for use in test_swig_import_simple


def test_swig_import_simple_BUGGY(tmp_path):
    # Test the currently implemented behavior of SWIG support.
    _test_swig_import_simple_common(tmp_path)


@xfail
def test_swig_import_simple(tmp_path):
    # Test the expected (but not implemented) behavior of SWIG support.
    mg = _test_swig_import_simple_common(tmp_path)
    # Given the bug in modulegraph (see test_swig_import_simple_BUGGY) this is what would be the expected behavior.
    # TODO: When modulegraph is fixed, merge the two test-cases and correct test_swig_import_from_top_level
    # and siblings.
    assert mg.find_node('pyi_test_osgeo._pyi_gdal') is None
    assert isinstance(mg.find_node('_pyi_gdal'), modulegraph.SourceModule)


def test_swig_import_from_top_level(tmp_path):
    # While there is a SWIG wrapper module as expected, the package module already imports the "C" module in the
    # same way the SWIG wrapper would do.
    # See the issue #1522 (at about 2017-04-26), pull-request #2578 and commit 711e9e77c93a979a63648ba05f725b30dbb7c3cc.
    #
    # For Python > 2.6, SWIG tries to import the C module from the package's directory and if this fails,
    # uses "import _XXX" (which is the code triggering import in modulegraph). For Python 2, this is a relative import,
    # but for Python 3, this is an absolute import.
    #
    # In this test-case, the package's __init__.py contains code equivalent to the SWIG wrapper-module, causing the C
    # module to be searched as an absolute import (in Python 3). But the importing module is not a SWIG candidate
    # (names do not match), leading to the (absolute) C module to become a MissingModule - which is okay up to this
    # point. Now if the SWIG wrapper-module imports the C module, there already is this MissingModule, inhibiting
    # modulegraph's SWIG import mechanism.
    #
    # This is where the commit 711e9e77c93 steps in and tries to reimport the C module (relative to the
    # SWIG wrapper-module).
    libdir = tmp_path / 'lib'

    osgeo = libdir / 'pyi_test_osgeo'
    osgeo.mkdir(parents=True)
    (osgeo / '__init__.py').write_text("import _pyi_gdal", encoding='utf-8')
    (osgeo / '_pyi_gdal.py').write_text("#", encoding='utf-8')
    (osgeo / 'pyi_gdal.py').write_text(
        "\n".join(["# automatically generated by SWIG", "import _pyi_gdal"]),
        encoding='utf-8',
    )

    script = tmp_path / 'script.py'
    script.write_text("from pyi_test_osgeo import pyi_gdal", encoding='utf-8')

    mg = modulegraph.ModuleGraph([str(libdir)])
    mg.add_script(str(script))

    assert isinstance(mg.find_node('pyi_test_osgeo'), modulegraph.Package)
    assert isinstance(mg.find_node('pyi_test_osgeo.pyi_gdal'), modulegraph.SourceModule)
    # The "C" module is frozen under its unqualified rather than qualified name.
    # See comment in modulegraph._safe_import_hook.
    # Due the the buggy implementation (see test_swig_import_simple):
    assert isinstance(mg.find_node('pyi_test_osgeo._pyi_gdal'), modulegraph.SourceModule)
    assert mg.find_node('_pyi_gdal') is None
    # This would be the correct implementation:
    #assert mg.find_node('pyi_test_osgeo._pyi_gdal') is None
    #assert isinstance(mg.find_node('_pyi_gdal'), modulegraph.SourceModule)


def test_swig_import_from_top_level_missing(tmp_path):
    # Like test_swig_import_from_top_level, but the "C" module is missing and should be reported as a MissingModule.
    libdir = tmp_path / 'lib'

    osgeo = libdir / 'pyi_test_osgeo'
    osgeo.mkdir(parents=True)
    (osgeo / '__init__.py').write_text("import _pyi_gdal", encoding='utf-8')
    (osgeo / 'pyi_gdal.py').write_text(
        "\n".join(["# automatically generated by SWIG", "import _pyi_gdal"]),
        encoding='utf-8',
    )

    script = tmp_path / 'script.py'
    script.write_text("from pyi_test_osgeo import pyi_gdal", encoding='utf-8')

    mg = modulegraph.ModuleGraph([str(libdir)])
    mg.add_script(str(script))

    assert isinstance(mg.find_node('pyi_test_osgeo'), modulegraph.Package)
    assert isinstance(mg.find_node('pyi_test_osgeo.pyi_gdal'), modulegraph.SourceModule)
    assert isinstance(mg.find_node('pyi_test_osgeo._pyi_gdal'), modulegraph.MissingModule)
    assert mg.find_node('_pyi_gdal') is None


def test_swig_import_from_top_level_but_nested(tmp_path):
    # Like test_swig_import_from_top_level, but both the wrapper and the "top level" are nested.
    # This is intended to test relative import of the "C" module.
    libdir = tmp_path / 'lib'

    osgeo = libdir / 'pyi_test_osgeo'
    (osgeo / 'x' / 'y').mkdir(parents=True)  # Create whole package directory tree.
    (osgeo / '__init__.py').write_text("#", encoding='utf-8')
    (osgeo / 'x' / '__init__.py').write_text("#", encoding='utf-8')
    (osgeo / 'x' / 'y' / '__init__.py').write_text("import _pyi_gdal", encoding='utf-8')
    (osgeo / 'x' / 'y' / '_pyi_gdal.py').write_text('#', encoding='utf-8')
    (osgeo / 'x' / 'y' / 'pyi_gdal.py').write_text(
        "\n".join(["# automatically generated by SWIG", "import _pyi_gdal"]),
        encoding='utf-8',
    )

    script = tmp_path / 'script.py'
    script.write_text("from pyi_test_osgeo.x.y import pyi_gdal", encoding='utf-8')

    mg = modulegraph.ModuleGraph([str(libdir)])
    mg.add_script(str(script))

    assert isinstance(mg.find_node('pyi_test_osgeo.x.y.pyi_gdal'), modulegraph.SourceModule)
    # The "C" module is frozen under its unqualified rather than qualified name.
    # See comment in modulegraph._safe_import_hook.
    # Due the the buggy implementation (see test_swig_import_simple):
    assert isinstance(mg.find_node('pyi_test_osgeo.x.y._pyi_gdal'), modulegraph.SourceModule)
    assert mg.find_node('_pyi_gdal') is None
    # This would be the correct implementation:
    #assert mg.find_node('pyi_test_osgeo.x.y._pyi_gdal') is None
    #assert isinstance(mg.find_node('_pyi_gdal'), modulegraph.SourceModule)


def test_swig_top_level_but_no_swig_at_all(tmp_path):
    # From the script import an absolute module which looks like a SWIG candidate but is no SWIG module.
    # See issue #3040 ('_decimal').
    # The center of this test-case is that it does not raise a recursion too deep error.
    libdir = tmp_path / 'lib'
    libdir.mkdir()
    (libdir / 'pyi_dezimal.py').write_text("import _pyi_dezimal", encoding='utf-8')

    script = tmp_path / 'script.py'
    script.write_text("import pyi_dezimal", encoding='utf-8')

    mg = modulegraph.ModuleGraph([str(libdir)])
    mg.add_script(str(script))

    assert isinstance(mg.find_node('pyi_dezimal'), modulegraph.SourceModule)
    assert isinstance(mg.find_node('_pyi_dezimal'), modulegraph.MissingModule)


def test_swig_top_level_but_no_swig_at_all_existing(tmp_path):
    # Like test_swig_top_level_but_no_swig_at_all, but the "C" module exists.
    # The test-case is here for symmetry.
    libdir = tmp_path / 'lib'
    libdir.mkdir()
    (libdir / 'pyi_dezimal.py').write_text("import _pyi_dezimal", encoding='utf-8')
    (libdir / '_pyi_dezimal.py').write_text("#", encoding='utf-8')

    script = tmp_path / 'script.py'
    script.write_text("import pyi_dezimal", encoding='utf-8')

    mg = modulegraph.ModuleGraph([str(libdir)])
    mg.add_script(str(script))

    assert isinstance(mg.find_node('pyi_dezimal'), modulegraph.SourceModule)
    assert isinstance(mg.find_node('_pyi_dezimal'), modulegraph.SourceModule)


def test_swig_candidate_but_not_swig(tmp_path):
    # From a package module import an absolute module which looks like a SWIG candidate, but is no SWIG module.
    # See issue #2911 (tifffile).
    # The center of this test-case is that it does not raise a recursion too deep error.
    libdir = tmp_path / 'lib'

    pkg = libdir / 'pkg'
    pkg.mkdir(parents=True)
    (pkg / '__init__.py').write_text("from . import mymod", encoding='utf-8')
    (pkg / 'mymod.py').write_text("import _mymod", encoding='utf-8')
    (pkg / '_mymod.py').write_text("#", encoding='utf-8')

    script = tmp_path / 'script.py'
    script.write_text("from pkg import XXX", encoding='utf-8')

    mg = modulegraph.ModuleGraph([str(libdir)])
    mg.add_script(str(script))

    assert isinstance(mg.find_node('pkg'), modulegraph.Package)
    assert isinstance(mg.find_node('pkg.mymod'), modulegraph.SourceModule)
    assert mg.find_node('pkg._mymod') is None
    # This is not a SWIG module, thus the SWIG import mechanism should not trigger.
    assert isinstance(mg.find_node('_mymod'), modulegraph.MissingModule)


def test_swig_candidate_but_not_swig2(tmp_path):
    """
    Variation of test_swig_candidate_but_not_swig using different import statements
    (like tifffile/tifffile.py does).
    """
    libdir = tmp_path / 'lib'

    pkg = libdir / 'pkg'
    pkg.mkdir(parents=True)
    (pkg / '__init__.py').write_text("from . import mymod", encoding='utf-8')
    (pkg / '_mymod.py').write_text("#", encoding='utf-8')
    (pkg / 'mymod.py').write_text(
        "\n".join(["from . import _mymod", "import _mymod"]),
        encoding='utf-8',
    )

    script = tmp_path / 'script.py'
    script.write_text("from pkg import XXX", encoding='utf-8')

    mg = modulegraph.ModuleGraph([str(libdir)])
    mg.add_script(str(script))

    assert isinstance(mg.find_node('pkg'), modulegraph.Package)
    assert isinstance(mg.find_node('pkg.mymod'), modulegraph.SourceModule)
    assert isinstance(mg.find_node('pkg._mymod'), modulegraph.SourceModule)
    assert isinstance(mg.find_node('_mymod'), modulegraph.MissingModule)


def test_swig_candidate_but_not_swig_missing(tmp_path):
    # Like test_swig_candidate_but_not_swig, but the "C" module is missing and should be reported as a MissingModule.
    libdir = tmp_path / 'lib'

    pkg = libdir / 'pkg'
    pkg.mkdir(parents=True)
    (pkg / '__init__.py').write_text("from . import mymod", encoding='utf-8')
    (pkg / 'mymod.py').write_text("import _mymod", encoding='utf-8')

    script = tmp_path / 'script.py'
    script.write_text("import pkg", encoding='utf-8')

    mg = modulegraph.ModuleGraph([str(libdir)])
    mg.add_script(str(script))

    assert isinstance(mg.find_node('pkg'), modulegraph.Package)
    assert isinstance(mg.find_node('pkg.mymod'), modulegraph.SourceModule)
    assert mg.find_node('pkg._mymod') is None
    assert isinstance(mg.find_node('_mymod'), modulegraph.MissingModule)


def test_swig_candidate_but_not_swig_missing2(tmp_path):
    """
    Variation of test_swig_candidate_but_not_swig_missing using different import statements
    (like tifffile/tifffile.py does).
    """
    libdir = tmp_path / 'lib'

    pkg = libdir / 'pkg'
    pkg.mkdir(parents=True)
    (pkg / '__init__.py').write_text("from . import mymod", encoding='utf-8')
    (pkg / 'mymod.py').write_text(
        "\n".join(["from . import _mymod", "import _mymod"]),
        encoding='utf-8',
    )

    script = tmp_path / 'script.py'
    script.write_text("import pkg", encoding='utf-8')

    mg = modulegraph.ModuleGraph([str(libdir)])
    mg.add_script(str(script))

    assert isinstance(mg.find_node('pkg'), modulegraph.Package)
    assert isinstance(mg.find_node('pkg.mymod'), modulegraph.SourceModule)
    assert isinstance(mg.find_node('pkg._mymod'), modulegraph.MissingModule)
    assert isinstance(mg.find_node('_mymod'), modulegraph.MissingModule)



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
GNU/Linux-specific test to check the bootloader from the AppImage.
"""

import os
import pathlib
import stat
import subprocess

import pytest


@pytest.mark.linux
@pytest.mark.parametrize('arch', ['x86_64'])
def test_appimage_loading(tmp_path, pyi_builder_spec, arch):
    # Skip the test if appimagetool is not found
    appimagetool = pathlib.Path.home() / f'appimagetool-{arch}.AppImage'
    if not appimagetool.is_file():
        pytest.skip(f'{str(appimagetool)!r} not found')

    # Ensure appimagetool is executable
    if not os.access(appimagetool, os.X_OK):
        st = appimagetool.stat()
        appimagetool.chmod(st.st_mode | stat.S_IXUSR)

    app_name = 'apptest'
    app_path = tmp_path / f'{app_name}-{arch}.AppImage'

    # Freeze the app
    pyi_builder_spec.test_source('print("OK")', app_name=app_name, pyi_args=["--onedir"])

    # Prepare the dist folder for AppImage compliance
    tools_dir = pathlib.Path(__file__).parent / 'data' / 'appimage'
    script = tools_dir / 'create.sh'
    subprocess.check_call(['bash', script, tools_dir, tmp_path, app_name])

    # Create the AppImage
    app_dir = tmp_path / 'dist' / 'AppRun'
    subprocess.check_call([appimagetool, "--no-appstream", app_dir, app_path])

    # Launch the AppImage
    st = app_path.stat()
    app_path.chmod(st.st_mode | stat.S_IXUSR)
    subprocess.check_call([app_path])



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

from waflib import Utils, Task, Options, Errors
from waflib.TaskGen import before_method, after_method, feature
from waflib.Tools import ccroot
from waflib.Configure import conf

ccroot.USELIB_VARS['cs'] = set(['CSFLAGS', 'ASSEMBLIES', 'RESOURCES'])
ccroot.lib_patterns['csshlib'] = ['%s']


@feature('cs')
@before_method('process_source')
def apply_cs(self):
    cs_nodes = []
    no_nodes = []
    for x in self.to_nodes(self.source):
        if x.name.endswith('.cs'):
            cs_nodes.append(x)
        else:
            no_nodes.append(x)
    self.source = no_nodes
    bintype = getattr(self, 'bintype', self.gen.endswith('.dll') and 'library' or 'exe')
    self.cs_task = tsk = self.create_task('mcs', cs_nodes, self.path.find_or_declare(self.gen))
    tsk.env.CSTYPE = '/target:%s' % bintype
    tsk.env.OUT = '/out:%s' % tsk.outputs[0].abspath()
    self.env.append_value('CSFLAGS', '/platform:%s' % getattr(self, 'platform', 'anycpu'))
    inst_to = getattr(self, 'install_path', bintype == 'exe' and '${BINDIR}' or '${LIBDIR}')
    if inst_to:
        mod = getattr(self, 'chmod', bintype == 'exe' and Utils.O755 or Utils.O644)
        self.install_task = self.add_install_files(install_to=inst_to, install_from=self.cs_task.outputs[:], chmod=mod)


@feature('cs')
@after_method('apply_cs')
def use_cs(self):
    names = self.to_list(getattr(self, 'use', []))
    get = self.bld.get_tgen_by_name
    for x in names:
        try:
            y = get(x)
        except Errors.WafError:
            self.env.append_value('CSFLAGS', '/reference:%s' % x)
            continue
        y.post()
        tsk = getattr(y, 'cs_task', None) or getattr(y, 'link_task', None)
        if not tsk:
            self.bld.fatal('cs task has no link task for use %r' % self)
        self.cs_task.dep_nodes.extend(tsk.outputs)
        self.cs_task.set_run_after(tsk)
        self.env.append_value('CSFLAGS', '/reference:%s' % tsk.outputs[0].abspath())


@feature('cs')
@after_method('apply_cs', 'use_cs')
def debug_cs(self):
    csdebug = getattr(self, 'csdebug', self.env.CSDEBUG)
    if not csdebug:
        return
    node = self.cs_task.outputs[0]
    if self.env.CS_NAME == 'mono':
        out = node.parent.find_or_declare(node.name + '.mdb')
    else:
        out = node.change_ext('.pdb')
    self.cs_task.outputs.append(out)
    if getattr(self, 'install_task', None):
        self.pdb_install_task = self.add_install_files(install_to=self.install_task.install_to, install_from=out)
    if csdebug == 'pdbonly':
        val = ['/debug+', '/debug:pdbonly']
    elif csdebug == 'full':
        val = ['/debug+', '/debug:full']
    else:
        val = ['/debug-']
    self.env.append_value('CSFLAGS', val)


@feature('cs')
@after_method('debug_cs')
def doc_cs(self):
    csdoc = getattr(self, 'csdoc', self.env.CSDOC)
    if not csdoc:
        return
    node = self.cs_task.outputs[0]
    out = node.change_ext('.xml')
    self.cs_task.outputs.append(out)
    if getattr(self, 'install_task', None):
        self.doc_install_task = self.add_install_files(install_to=self.install_task.install_to, install_from=out)
    self.env.append_value('CSFLAGS', '/doc:%s' % out.abspath())


class mcs(Task.Task):
    color = 'YELLOW'
    run_str = '${MCS} ${CSTYPE} ${CSFLAGS} ${ASS_ST:ASSEMBLIES} ${RES_ST:RESOURCES} ${OUT} ${SRC}'

    def split_argfile(self, cmd):
        inline = [cmd[0]]
        infile = []
        for x in cmd[1:]:
            if x.lower() == '/noconfig':
                inline.append(x)
            else:
                infile.append(self.quote_flag(x))
        return (inline, infile)


def configure(conf):
    csc = getattr(Options.options, 'cscbinary', None)
    if csc:
        conf.env.MCS = csc
    conf.find_program(['csc', 'mcs', 'gmcs'], var='MCS')
    conf.env.ASS_ST = '/r:%s'
    conf.env.RES_ST = '/resource:%s'
    conf.env.CS_NAME = 'csc'
    if str(conf.env.MCS).lower().find('mcs') > -1:
        conf.env.CS_NAME = 'mono'


def options(opt):
    opt.add_option('--with-csc-binary', type='string', dest='cscbinary')


class fake_csshlib(Task.Task):
    color = 'YELLOW'
    inst_to = None

    def runnable_status(self):
        return Task.SKIP_ME


@conf
def read_csshlib(self, name, paths=[]):
    return self(name=name, features='fake_lib', lib_paths=paths, lib_type='csshlib')



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

from waflib import Task, Errors
from waflib.TaskGen import taskgen_method, before_method


@taskgen_method
def add_dbus_file(self, filename, prefix, mode):
    if not hasattr(self, 'dbus_lst'):
        self.dbus_lst = []
    if not 'process_dbus' in self.meths:
        self.meths.append('process_dbus')
    self.dbus_lst.append([filename, prefix, mode])


@before_method('process_source')
def process_dbus(self):
    for filename, prefix, mode in getattr(self, 'dbus_lst', []):
        node = self.path.find_resource(filename)
        if not node:
            raise Errors.WafError('file not found ' + filename)
        tsk = self.create_task('dbus_binding_tool', node, node.change_ext('.h'))
        tsk.env.DBUS_BINDING_TOOL_PREFIX = prefix
        tsk.env.DBUS_BINDING_TOOL_MODE = mode


class dbus_binding_tool(Task.Task):
    color = 'BLUE'
    ext_out = ['.h']
    run_str = '${DBUS_BINDING_TOOL} --prefix=${DBUS_BINDING_TOOL_PREFIX} --mode=${DBUS_BINDING_TOOL_MODE} --output=${TGT} ${SRC}'
    shell = True


def configure(conf):
    conf.find_program('dbus-binding-tool', var='DBUS_BINDING_TOOL')



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import copy, re, os, functools
from waflib import Task, Utils, Logs, Errors, ConfigSet, Node

feats = Utils.defaultdict(set)
HEADER_EXTS = ['.h', '.hpp', '.hxx', '.hh']


class task_gen(object):
    mappings = Utils.ordered_iter_dict()
    prec = Utils.defaultdict(set)

    def __init__(self, *k, **kw):
        self.source = []
        self.target = ''
        self.meths = []
        self.features = []
        self.tasks = []
        if not 'bld' in kw:
            self.env = ConfigSet.ConfigSet()
            self.idx = 0
            self.path = None
        else:
            self.bld = kw['bld']
            self.env = self.bld.env.derive()
            self.path = kw.get('path', self.bld.path)
            path = self.path.abspath()
            try:
                self.idx = self.bld.idx[path] = self.bld.idx.get(path, 0) + 1
            except AttributeError:
                self.bld.idx = {}
                self.idx = self.bld.idx[path] = 1
            try:
                self.tg_idx_count = self.bld.tg_idx_count = self.bld.tg_idx_count + 1
            except AttributeError:
                self.tg_idx_count = self.bld.tg_idx_count = 1
        for key, val in kw.items():
            setattr(self, key, val)

    def __str__(self):
        return "<task_gen %r declared in %s>" % (self.name, self.path.abspath())

    def __repr__(self):
        lst = []
        for x in self.__dict__:
            if x not in ('env', 'bld', 'compiled_tasks', 'tasks'):
                lst.append("%s=%s" % (x, repr(getattr(self, x))))
        return "bld(%s) in %s" % (", ".join(lst), self.path.abspath())

    def get_cwd(self):
        return self.bld.bldnode

    def get_name(self):
        try:
            return self._name
        except AttributeError:
            if isinstance(self.target, list):
                lst = [str(x) for x in self.target]
                name = self._name = ','.join(lst)
            else:
                name = self._name = str(self.target)
            return name

    def set_name(self, name):
        self._name = name

    name = property(get_name, set_name)

    def to_list(self, val):
        if isinstance(val, str):
            return val.split()
        else:
            return val

    def post(self):
        if getattr(self, 'posted', None):
            return False
        self.posted = True
        keys = set(self.meths)
        keys.update(feats['*'])
        self.features = Utils.to_list(self.features)
        for x in self.features:
            st = feats[x]
            if st:
                keys.update(st)
            elif not x in Task.classes:
                Logs.warn('feature %r does not exist - bind at least one method to it?', x)
        prec = {}
        prec_tbl = self.prec
        for x in prec_tbl:
            if x in keys:
                prec[x] = prec_tbl[x]
        tmp = []
        for a in keys:
            for x in prec.values():
                if a in x:
                    break
            else:
                tmp.append(a)
        tmp.sort(reverse=True)
        out = []
        while tmp:
            e = tmp.pop()
            if e in keys:
                out.append(e)
            try:
                nlst = prec[e]
            except KeyError:
                pass
            else:
                del prec[e]
                for x in nlst:
                    for y in prec:
                        if x in prec[y]:
                            break
                    else:
                        tmp.append(x)
                        tmp.sort(reverse=True)
        if prec:
            buf = ['Cycle detected in the method execution:']
            for k, v in prec.items():
                buf.append('- %s after %s' % (k, [x for x in v if x in prec]))
            raise Errors.WafError('\n'.join(buf))
        self.meths = out
        Logs.debug('task_gen: posting %s %d', self, id(self))
        for x in out:
            try:
                v = getattr(self, x)
            except AttributeError:
                raise Errors.WafError('%r is not a valid task generator method' % x)
            Logs.debug('task_gen: -> %s (%d)', x, id(self))
            v()
        Logs.debug('task_gen: posted %s', self.name)
        return True

    def get_hook(self, node):
        name = node.name
        for k in self.mappings:
            try:
                if name.endswith(k):
                    return self.mappings[k]
            except TypeError:
                if k.match(name):
                    return self.mappings[k]
        keys = list(self.mappings.keys())
        raise Errors.WafError("File %r has no mapping in %r (load a waf tool?)" % (node, keys))

    def create_task(self, name, src=None, tgt=None, **kw):
        task = Task.classes[name](env=self.env.derive(), generator=self)
        if src:
            task.set_inputs(src)
        if tgt:
            task.set_outputs(tgt)
        task.__dict__.update(kw)
        self.tasks.append(task)
        return task

    def clone(self, env):
        newobj = self.bld()
        for x in self.__dict__:
            if x in ('env', 'bld'):
                continue
            elif x in ('path', 'features'):
                setattr(newobj, x, getattr(self, x))
            else:
                setattr(newobj, x, copy.copy(getattr(self, x)))
        newobj.posted = False
        if isinstance(env, str):
            newobj.env = self.bld.all_envs[env].derive()
        else:
            newobj.env = env.derive()
        return newobj


def declare_chain(
    name='',
    rule=None,
    reentrant=None,
    color='BLUE',
    ext_in=[],
    ext_out=[],
    before=[],
    after=[],
    decider=None,
    scan=None,
    install_path=None,
    shell=False
):
    ext_in = Utils.to_list(ext_in)
    ext_out = Utils.to_list(ext_out)
    if not name:
        name = rule
    cls = Task.task_factory(
        name, rule, color=color, ext_in=ext_in, ext_out=ext_out, before=before, after=after, scan=scan, shell=shell
    )

    def x_file(self, node):
        if ext_in:
            _ext_in = ext_in[0]
        tsk = self.create_task(name, node)
        cnt = 0
        ext = decider(self, node) if decider else cls.ext_out
        for x in ext:
            k = node.change_ext(x, ext_in=_ext_in)
            tsk.outputs.append(k)
            if reentrant != None:
                if cnt < int(reentrant):
                    self.source.append(k)
            else:
                for y in self.mappings:
                    if k.name.endswith(y):
                        self.source.append(k)
                        break
            cnt += 1
        if install_path:
            self.install_task = self.add_install_files(install_to=install_path, install_from=tsk.outputs)
        return tsk

    for x in cls.ext_in:
        task_gen.mappings[x] = x_file
    return x_file


def taskgen_method(func):
    setattr(task_gen, func.__name__, func)
    return func


def feature(*k):
    def deco(func):
        setattr(task_gen, func.__name__, func)
        for name in k:
            feats[name].update([func.__name__])
        return func

    return deco


def before_method(*k):
    def deco(func):
        setattr(task_gen, func.__name__, func)
        for fun_name in k:
            task_gen.prec[func.__name__].add(fun_name)
        return func

    return deco


before = before_method


def after_method(*k):
    def deco(func):
        setattr(task_gen, func.__name__, func)
        for fun_name in k:
            task_gen.prec[fun_name].add(func.__name__)
        return func

    return deco


after = after_method


def extension(*k):
    def deco(func):
        setattr(task_gen, func.__name__, func)
        for x in k:
            task_gen.mappings[x] = func
        return func

    return deco


@taskgen_method
def to_nodes(self, lst, path=None):
    tmp = []
    path = path or self.path
    find = path.find_resource
    if isinstance(lst, Node.Node):
        lst = [lst]
    for x in Utils.to_list(lst):
        if isinstance(x, str):
            node = find(x)
        elif hasattr(x, 'name'):
            node = x
        else:
            tmp.extend(self.to_nodes(x))
            continue
        if not node:
            raise Errors.WafError('source not found: %r in %r' % (x, self))
        tmp.append(node)
    return tmp


@feature('*')
def process_source(self):
    self.source = self.to_nodes(getattr(self, 'source', []))
    for node in self.source:
        self.get_hook(node)(self, node)


@feature('*')
@before_method('process_source')
def process_rule(self):
    if not getattr(self, 'rule', None):
        return
    name = str(getattr(self, 'name', None) or self.target or getattr(self.rule, '__name__', self.rule))
    try:
        cache = self.bld.cache_rule_attr
    except AttributeError:
        cache = self.bld.cache_rule_attr = {}
    chmod = getattr(self, 'chmod', None)
    shell = getattr(self, 'shell', True)
    color = getattr(self, 'color', 'BLUE')
    scan = getattr(self, 'scan', None)
    _vars = getattr(self, 'vars', [])
    cls_str = getattr(self, 'cls_str', None)
    cls_keyword = getattr(self, 'cls_keyword', None)
    use_cache = getattr(self, 'cache_rule', 'True')
    deep_inputs = getattr(self, 'deep_inputs', False)
    scan_val = has_deps = hasattr(self, 'deps')
    if scan:
        scan_val = id(scan)
    key = Utils.h_list((name, self.rule, chmod, shell, color, cls_str, cls_keyword, scan_val, _vars, deep_inputs))
    cls = None
    if use_cache:
        try:
            cls = cache[key]
        except KeyError:
            pass
    if not cls:
        rule = self.rule
        if chmod is not None:

            def chmod_fun(tsk):
                for x in tsk.outputs:
                    os.chmod(x.abspath(), tsk.generator.chmod)

            if isinstance(rule, tuple):
                rule = list(rule)
                rule.append(chmod_fun)
                rule = tuple(rule)
            else:
                rule = (rule, chmod_fun)
        cls = Task.task_factory(name, rule, _vars, shell=shell, color=color)
        if cls_str:
            setattr(cls, '__str__', self.cls_str)
        if cls_keyword:
            setattr(cls, 'keyword', self.cls_keyword)
        if deep_inputs:
            Task.deep_inputs(cls)
        if scan:
            cls.scan = self.scan
        elif has_deps:

            def scan(self):
                nodes = []
                for x in self.generator.to_list(getattr(self.generator, 'deps', None)):
                    node = self.generator.path.find_resource(x)
                    if not node:
                        self.generator.bld.fatal('Could not find %r (was it declared?)' % x)
                    nodes.append(node)
                return [nodes, []]

            cls.scan = scan
        if use_cache:
            cache[key] = cls
    tsk = self.create_task(name)
    for x in ('after', 'before', 'ext_in', 'ext_out'):
        setattr(tsk, x, getattr(self, x, []))
    if hasattr(self, 'stdout'):
        tsk.stdout = self.stdout
    if hasattr(self, 'stderr'):
        tsk.stderr = self.stderr
    if getattr(self, 'timeout', None):
        tsk.timeout = self.timeout
    if getattr(self, 'always', None):
        tsk.always_run = True
    if getattr(self, 'target', None):
        if isinstance(self.target, str):
            self.target = self.target.split()
        if not isinstance(self.target, list):
            self.target = [self.target]
        for x in self.target:
            if isinstance(x, str):
                tsk.outputs.append(self.path.find_or_declare(x))
            else:
                x.parent.mkdir()
                tsk.outputs.append(x)
        if getattr(self, 'install_path', None):
            self.install_task = self.add_install_files(
                install_to=self.install_path, install_from=tsk.outputs, chmod=getattr(self, 'chmod', Utils.O644)
            )
    if getattr(self, 'source', None):
        tsk.inputs = self.to_nodes(self.source)
        self.source = []
    if getattr(self, 'cwd', None):
        tsk.cwd = self.cwd
    if isinstance(tsk.run, functools.partial):
        tsk.run = functools.partial(tsk.run, tsk)


@feature('seq')
def sequence_order(self):
    if self.meths and self.meths[-1] != 'sequence_order':
        self.meths.append('sequence_order')
        return
    if getattr(self, 'seq_start', None):
        return
    if getattr(self.bld, 'prev', None):
        self.bld.prev.post()
        for x in self.bld.prev.tasks:
            for y in self.tasks:
                y.set_run_after(x)
    self.bld.prev = self


re_m4 = re.compile(r'@(\w+)@', re.M)


class subst_pc(Task.Task):
    def force_permissions(self):
        if getattr(self.generator, 'chmod', None):
            for x in self.outputs:
                os.chmod(x.abspath(), self.generator.chmod)

    def run(self):
        if getattr(self.generator, 'is_copy', None):
            for i, x in enumerate(self.outputs):
                x.write(self.inputs[i].read('rb'), 'wb')
                stat = os.stat(self.inputs[i].abspath())
                os.utime(self.outputs[i].abspath(), (stat.st_atime, stat.st_mtime))
            self.force_permissions()
            return None
        if getattr(self.generator, 'fun', None):
            ret = self.generator.fun(self)
            if not ret:
                self.force_permissions()
            return ret
        code = self.inputs[0].read(encoding=getattr(self.generator, 'encoding', 'latin-1'))
        if getattr(self.generator, 'subst_fun', None):
            code = self.generator.subst_fun(self, code)
            if code is not None:
                self.outputs[0].write(code, encoding=getattr(self.generator, 'encoding', 'latin-1'))
            self.force_permissions()
            return None
        code = code.replace('%', '%%')
        lst = []

        def repl(match):
            g = match.group
            if g(1):
                lst.append(g(1))
                return "%%(%s)s" % g(1)
            return ''

        code = getattr(self.generator, 're_m4', re_m4).sub(repl, code)
        try:
            d = self.generator.dct
        except AttributeError:
            d = {}
            for x in lst:
                tmp = getattr(self.generator, x, '') or self.env[x] or self.env[x.upper()]
                try:
                    tmp = ''.join(tmp)
                except TypeError:
                    tmp = str(tmp)
                d[x] = tmp
        code = code % d
        self.outputs[0].write(code, encoding=getattr(self.generator, 'encoding', 'latin-1'))
        self.generator.bld.raw_deps[self.uid()] = lst
        try:
            delattr(self, 'cache_sig')
        except AttributeError:
            pass
        self.force_permissions()

    def sig_vars(self):
        bld = self.generator.bld
        env = self.env
        upd = self.m.update
        if getattr(self.generator, 'fun', None):
            upd(Utils.h_fun(self.generator.fun).encode())
        if getattr(self.generator, 'subst_fun', None):
            upd(Utils.h_fun(self.generator.subst_fun).encode())
        vars = self.generator.bld.raw_deps.get(self.uid(), [])
        act_sig = bld.hash_env_vars(env, vars)
        upd(act_sig)
        lst = [getattr(self.generator, x, '') for x in vars]
        upd(Utils.h_list(lst))
        return self.m.digest()


@extension('.pc.in')
def add_pcfile(self, node):
    tsk = self.create_task('subst_pc', node, node.change_ext('.pc', '.pc.in'))
    self.install_task = self.add_install_files(
        install_to=getattr(self, 'install_path', '${LIBDIR}/pkgconfig/'), install_from=tsk.outputs
    )


class subst(subst_pc):
    pass


@feature('subst')
@before_method('process_source', 'process_rule')
def process_subst(self):
    src = Utils.to_list(getattr(self, 'source', []))
    if isinstance(src, Node.Node):
        src = [src]
    tgt = Utils.to_list(getattr(self, 'target', []))
    if isinstance(tgt, Node.Node):
        tgt = [tgt]
    if len(src) != len(tgt):
        raise Errors.WafError('invalid number of source/target for %r' % self)
    for x, y in zip(src, tgt):
        if not x or not y:
            raise Errors.WafError('null source or target for %r' % self)
        a, b = None, None
        if isinstance(x, str) and isinstance(y, str) and x == y:
            a = self.path.find_node(x)
            b = self.path.get_bld().make_node(y)
            if not os.path.isfile(b.abspath()):
                b.parent.mkdir()
        else:
            if isinstance(x, str):
                a = self.path.find_resource(x)
            elif isinstance(x, Node.Node):
                a = x
            if isinstance(y, str):
                b = self.path.find_or_declare(y)
            elif isinstance(y, Node.Node):
                b = y
        if not a:
            raise Errors.WafError('could not find %r for %r' % (x, self))
        tsk = self.create_task('subst', a, b)
        for k in ('after', 'before', 'ext_in', 'ext_out'):
            val = getattr(self, k, None)
            if val:
                setattr(tsk, k, val)
        for xt in HEADER_EXTS:
            if b.name.endswith(xt):
                tsk.ext_out = tsk.ext_out + ['.h']
                break
        inst_to = getattr(self, 'install_path', None)
        if inst_to:
            self.install_task = self.add_install_files(
                install_to=inst_to, install_from=b, chmod=getattr(self, 'chmod', Utils.O644)
            )
    self.source = []



"""
Tests that deal with setuptools namespace
packages, and in particular the installation
flavour used by pip
"""
import os
import sys
import subprocess
import unittest
import textwrap
import shutil

import pytest

from PyInstaller.lib.modulegraph import modulegraph

gRootDir = os.path.dirname(os.path.abspath(__file__))
gSrcDir = os.path.join(gRootDir, 'testpkg-setuptools-namespace')


@pytest.fixture
def install_testpkg(tmp_path):
    # Copy the package to ``dest_dir``, so that the build won't modify anything in ``gSrcDir``.
    dest_dir = tmp_path / 'data'
    shutil.copytree(gSrcDir, dest_dir)

    # A directory to place the resulting built library in.
    libdir = tmp_path / 'test'

    # Perform the build.
    subprocess.check_call([
        sys.executable, 'setup.py', 'install',
            '--install-lib', str(libdir),
            '--single-version-externally-managed',
            '--record', str(libdir  / 'record.lst'),
        ], cwd=dest_dir)

    return libdir


class TestPythonBehaviour(object):

    def importModule(self, name, libdir):
        if '.' in name:
            script = textwrap.dedent("""\
                import site
                site.addsitedir(%r)
                try:
                    import %s
                except ImportError:
                    import %s
                print (%s.__name__)
            """) % (str(libdir), name, name.rsplit('.', 1)[0], name)
        else:
            script = textwrap.dedent("""\
                import site
                site.addsitedir(%r)
                import %s
                print (%s.__name__)
            """) % (str(libdir), name, name)

        data = subprocess.check_output(
            [sys.executable, '-c', script],
            stderr=subprocess.STDOUT,
            cwd=os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                'testpkg-relimport'),
            encoding='utf8',
        )
        data = data.strip()
        if data.endswith(' refs]'):
            data = data.rsplit('\n', 1)[0].strip()

        return data

    def testToplevel(self, install_testpkg):
        m = self.importModule('nspkg.module', install_testpkg)
        assert m == 'nspkg.module'

    def testSub(self, install_testpkg):
        m = self.importModule('nspkg.nssubpkg.sub', install_testpkg)
        assert m == 'nspkg.nssubpkg.sub'


@pytest.fixture
def install_testpkg_modulegraph(install_testpkg):
    return modulegraph.ModuleGraph(path=[str(install_testpkg)] + sys.path)


class TestModuleGraphImport(object):

    def testRootPkg(self, install_testpkg_modulegraph):
        install_testpkg_modulegraph.import_hook('nspkg')

        node = install_testpkg_modulegraph.find_node('nspkg')
        assert isinstance(node, modulegraph.NamespacePackage)
        assert node.identifier == 'nspkg'
        assert node.filename == '-'

    def testRootPkgModule(self, install_testpkg_modulegraph):
        install_testpkg_modulegraph.import_hook('nspkg.module')

        node = install_testpkg_modulegraph.find_node('nspkg.module')
        assert isinstance(node, modulegraph.SourceModule)
        assert node.identifier == 'nspkg.module'

    def testSubRootPkgModule(self, install_testpkg_modulegraph):
        install_testpkg_modulegraph.import_hook('nspkg.nssubpkg.sub')

        node = install_testpkg_modulegraph.find_node('nspkg.nssubpkg.sub')
        assert isinstance(node, modulegraph.SourceModule)
        assert node.identifier == 'nspkg.nssubpkg.sub'

        node = install_testpkg_modulegraph.find_node('nspkg')
        assert isinstance(node, modulegraph.NamespacePackage)


if __name__ == "__main__":
    unittest.main()



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import os, sys, traceback, base64, signal
try:
    import cPickle
except ImportError:
    import pickle as cPickle
try:
    import subprocess32 as subprocess
except ImportError:
    import subprocess
try:
    TimeoutExpired = subprocess.TimeoutExpired
except AttributeError:

    class TimeoutExpired(Exception):
        pass


def run():
    txt = sys.stdin.readline().strip()
    if not txt:
        sys.exit(1)
    [cmd, kwargs, cargs] = cPickle.loads(base64.b64decode(txt))
    cargs = cargs or {}
    if not 'close_fds' in kwargs:
        kwargs['close_fds'] = False
    ret = 1
    out, err, ex, trace = (None, None, None, None)
    try:
        proc = subprocess.Popen(cmd, **kwargs)
        try:
            out, err = proc.communicate(**cargs)
        except TimeoutExpired:
            if kwargs.get('start_new_session') and hasattr(os, 'killpg'):
                os.killpg(proc.pid, signal.SIGKILL)
            else:
                proc.kill()
            out, err = proc.communicate()
            exc = TimeoutExpired(proc.args, timeout=cargs['timeout'], output=out)
            exc.stderr = err
            raise exc
        ret = proc.returncode
    except Exception as e:
        exc_type, exc_value, tb = sys.exc_info()
        exc_lines = traceback.format_exception(exc_type, exc_value, tb)
        trace = str(cmd) + '\n' + ''.join(exc_lines)
        ex = e.__class__.__name__
    tmp = [ret, out, err, ex, trace]
    obj = base64.b64encode(cPickle.dumps(tmp))
    sys.stdout.write(obj.decode())
    sys.stdout.write('\n')
    sys.stdout.flush()


while 1:
    try:
        run()
    except KeyboardInterrupt:
        break



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

from waflib import Utils, Errors
from waflib.Configure import conf


def get_extensions(lst):
    ret = []
    for x in Utils.to_list(lst):
        if not isinstance(x, str):
            x = x.name
        ret.append(x[x.rfind('.') + 1:])
    return ret


def sniff_features(**kw):
    exts = get_extensions(kw.get('source', []))
    typ = kw['typ']
    feats = []
    for x in 'cxx cpp c++ cc C'.split():
        if x in exts:
            feats.append('cxx')
            break
    if 'c' in exts or 'vala' in exts or 'gs' in exts:
        feats.append('c')
    if 's' in exts or 'S' in exts:
        feats.append('asm')
    for x in 'f f90 F F90 for FOR'.split():
        if x in exts:
            feats.append('fc')
            break
    if 'd' in exts:
        feats.append('d')
    if 'java' in exts:
        feats.append('java')
        return 'java'
    if typ in ('program', 'shlib', 'stlib'):
        will_link = False
        for x in feats:
            if x in ('cxx', 'd', 'fc', 'c', 'asm'):
                feats.append(x + typ)
                will_link = True
        if not will_link and not kw.get('features', []):
            raise Errors.WafError('Unable to determine how to link %r, try adding eg: features="c cshlib"?' % kw)
    return feats


def set_features(kw, typ):
    kw['typ'] = typ
    kw['features'] = Utils.to_list(kw.get('features', [])) + Utils.to_list(sniff_features(**kw))


@conf
def program(bld, *k, **kw):
    set_features(kw, 'program')
    return bld(*k, **kw)


@conf
def shlib(bld, *k, **kw):
    set_features(kw, 'shlib')
    return bld(*k, **kw)


@conf
def stlib(bld, *k, **kw):
    set_features(kw, 'stlib')
    return bld(*k, **kw)


@conf
def objects(bld, *k, **kw):
    set_features(kw, 'objects')
    return bld(*k, **kw)



#-----------------------------------------------------------------------------
# Copyright (c) 2014-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

from PyInstaller.utils.hooks.qt import \
    add_qt5_dependencies, pyqt5_library_info

# Ensure PyQt5 is importable before adding info depending on it.
if pyqt5_library_info.version is not None:
    hiddenimports, binaries, datas = add_qt5_dependencies(__file__)

    # Include helper process executable, translations, and resources.
    webengine_binaries, webengine_datas = pyqt5_library_info.collect_qtwebengine_files()
    binaries += webengine_binaries
    datas += webengine_datas



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
Interactive tests are successful when they are able to run the executable for some time.
Otherwise it is marked as fail.

Note: All tests in this file should use the argument 'runtime'.
"""

from PyInstaller.utils.tests import importorskip

_RUNTIME = 10  # In seconds.


@importorskip('IPython')
def test_ipython(pyi_builder):
    pyi_builder.test_source("""
        from IPython import embed
        embed()
        """, runtime=_RUNTIME)



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License with exception
# for distributing bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------


def pre_safe_import_module(api):
    api.add_alias_module('pyi_testmod_submodule_from_aliased_pkg', 'alias_name')



#-----------------------------------------------------------------------------
# Copyright (c) 2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

from PyInstaller.utils.hooks.gi import GiModuleInfo


def test_module_info():
    module_info = GiModuleInfo("MyModule", "1.0")

    assert module_info.name == "MyModule"
    assert module_info.version == "1.0"


def test_module_info_with_versions():
    hook_api = hook_api_for_module_version("MyModule", "2.0")

    module_info = GiModuleInfo("MyModule", "1.0", hook_api)

    assert module_info.name == "MyModule"
    assert module_info.version == "2.0"


def hook_api_for_module_version(module, version):
    class hook_api_stub:
        class analysis:
            hooksconfig = {
                "gi": {
                    "module-versions": {
                        module: version,
                    }
                }
            }

    return hook_api_stub



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
import xml.dom
import xml.dom.minidom

#- Relevant constants from Windows headers
# Manifest resource code
RT_MANIFEST = 24

# Resource IDs (names) for manifest.
# See: https://www.gamedev.net/blogs/entry/2154553-manifest-embedding-and-activation
CREATEPROCESS_MANIFEST_RESOURCE_ID = 1
ISOLATIONAWARE_MANIFEST_RESOURCE_ID = 2

LANG_NEUTRAL = 0

#- Default application manifest template, based on the one found in python executable.

_DEFAULT_MANIFEST_XML = \
b"""<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<assembly xmlns="urn:schemas-microsoft-com:asm.v1" manifestVersion="1.0">
  <trustInfo xmlns="urn:schemas-microsoft-com:asm.v3">
    <security>
      <requestedPrivileges>
        <requestedExecutionLevel level="asInvoker" uiAccess="false"></requestedExecutionLevel>
      </requestedPrivileges>
    </security>
  </trustInfo>
  <compatibility xmlns="urn:schemas-microsoft-com:compatibility.v1">
    <application>
      <supportedOS Id="{e2011457-1546-43c5-a5fe-008deee3d3f0}"></supportedOS>
      <supportedOS Id="{35138b9a-5d96-4fbd-8e2d-a2440225f93a}"></supportedOS>
      <supportedOS Id="{4a2f28e3-53b9-4441-ba9c-d69d4a4a6e38}"></supportedOS>
      <supportedOS Id="{1f676c76-80e1-4239-95bb-83d0f6d0da78}"></supportedOS>
      <supportedOS Id="{8e0f7a12-bfb3-4fe8-b9a5-48fd50a15a9a}"></supportedOS>
    </application>
  </compatibility>
  <application xmlns="urn:schemas-microsoft-com:asm.v3">
    <windowsSettings>
      <longPathAware xmlns="http://schemas.microsoft.com/SMI/2016/WindowsSettings">true</longPathAware>
    </windowsSettings>
  </application>
  <dependency>
    <dependentAssembly>
      <assemblyIdentity type="win32" name="Microsoft.Windows.Common-Controls" version="6.0.0.0" processorArchitecture="*" publicKeyToken="6595b64144ccf1df" language="*"></assemblyIdentity>
    </dependentAssembly>
  </dependency>
</assembly>
"""  # noqa: E122,E501

#- DOM navigation helpers


def _find_elements_by_tag(root, tag):
    """
    Find all elements with given tag under the given root element.
    """
    return [node for node in root.childNodes if node.nodeType == xml.dom.Node.ELEMENT_NODE and node.tagName == tag]


def _find_element_by_tag(root, tag):
    """
    Attempt to find a single element with given tag under the given root element, and return None if no such element
    is found. Raises an error if multiple elements are found.
    """
    elements = _find_elements_by_tag(root, tag)
    if len(elements) > 1:
        raise ValueError(f"Expected a single {tag!r} element, found {len(elements)} element(s)!")
    if not elements:
        return None
    return elements[0]


#- Application manifest modification helpers


def _set_execution_level(manifest_dom, root_element, uac_admin=False, uac_uiaccess=False):
    """
    Find <security> -> <requestedPrivileges> -> <requestedExecutionLevel> element, and set its `level` and `uiAccess`
    attributes based on supplied arguments. Create the XML elements if necessary, as they are optional.
    """

    # <trustInfo xmlns="urn:schemas-microsoft-com:asm.v3">
    trust_info_element = _find_element_by_tag(root_element, "trustInfo")
    if not trust_info_element:
        trust_info_element = manifest_dom.createElement("trustInfo")
        trust_info_element.setAttribute("xmlns", "urn:schemas-microsoft-com:asm.v3")
        root_element.appendChild(trust_info_element)

    # <security>
    security_element = _find_element_by_tag(trust_info_element, "security")
    if not security_element:
        security_element = manifest_dom.createElement("security")
        trust_info_element.appendChild(security_element)

    # <requestedPrivileges>
    requested_privileges_element = _find_element_by_tag(security_element, "requestedPrivileges")
    if not requested_privileges_element:
        requested_privileges_element = manifest_dom.createElement("requestedPrivileges")
        security_element.appendChild(requested_privileges_element)

    # <requestedExecutionLevel>
    requested_execution_level_element = _find_element_by_tag(requested_privileges_element, "requestedExecutionLevel")
    if not requested_execution_level_element:
        requested_execution_level_element = manifest_dom.createElement("requestedExecutionLevel")
        requested_privileges_element.appendChild(requested_execution_level_element)

    requested_execution_level_element.setAttribute("level", "requireAdministrator" if uac_admin else "asInvoker")
    requested_execution_level_element.setAttribute("uiAccess", "true" if uac_uiaccess else "false")


def _ensure_common_controls_dependency(manifest_dom, root_element):
    """
    Scan <dependency> elements for the one whose <<dependentAssembly> -> <assemblyIdentity> corresponds to the
    `Microsoft.Windows.Common-Controls`. If found, overwrite its properties. If not, create new <dependency>
    element with corresponding sub-elements and attributes.
    """

    # <dependency>
    dependency_elements = _find_elements_by_tag(root_element, "dependency")
    for dependency_element in dependency_elements:
        # <dependentAssembly>
        dependent_assembly_element = _find_element_by_tag(dependency_element, "dependentAssembly")
        # <assemblyIdentity>
        assembly_identity_element = _find_element_by_tag(dependent_assembly_element, "assemblyIdentity")
        # Check the name attribute
        if assembly_identity_element.attributes["name"].value == "Microsoft.Windows.Common-Controls":
            common_controls_element = assembly_identity_element
            break
    else:
        # Create <dependency>
        dependency_element = manifest_dom.createElement("dependency")
        root_element.appendChild(dependency_element)
        # Create <dependentAssembly>
        dependent_assembly_element = manifest_dom.createElement("dependentAssembly")
        dependency_element.appendChild(dependent_assembly_element)
        # Create <assemblyIdentity>
        common_controls_element = manifest_dom.createElement("assemblyIdentity")
        dependent_assembly_element.appendChild(common_controls_element)

    common_controls_element.setAttribute("type", "win32")
    common_controls_element.setAttribute("name", "Microsoft.Windows.Common-Controls")
    common_controls_element.setAttribute("version", "6.0.0.0")
    common_controls_element.setAttribute("processorArchitecture", "*")
    common_controls_element.setAttribute("publicKeyToken", "6595b64144ccf1df")
    common_controls_element.setAttribute("language", "*")


def create_application_manifest(manifest_xml=None, uac_admin=False, uac_uiaccess=False):
    """
    Create application manifest, from built-in or custom manifest XML template. If provided, `manifest_xml` must be
    a string or byte string containing XML source. The returned manifest is a byte string, encoded in UTF-8.

    This function sets the attributes of `requestedExecutionLevel` based on provided `uac_admin` and `auc_uiacces`
    arguments (creating the parent elements in the XML, if necessary). It also scans `dependency` elements for the
    entry corresponding to `Microsoft.Windows.Common-Controls` and creates or modifies it as necessary.
    """

    if manifest_xml is None:
        manifest_xml = _DEFAULT_MANIFEST_XML

    with xml.dom.minidom.parseString(manifest_xml) as manifest_dom:
        root_element = manifest_dom.documentElement

        # Validate root element - must be <assembly>
        assert root_element.tagName == "assembly"
        assert root_element.namespaceURI == "urn:schemas-microsoft-com:asm.v1"
        assert root_element.attributes["manifestVersion"].value == "1.0"

        # Modify the manifest
        _set_execution_level(manifest_dom, root_element, uac_admin, uac_uiaccess)
        _ensure_common_controls_dependency(manifest_dom, root_element)

        # Create output XML
        output = manifest_dom.toprettyxml(indent="  ", encoding="UTF-8")

    # Strip extra newlines
    output = [line for line in output.splitlines() if line.strip()]

    # Replace: `<?xml version="1.0" encoding="UTF-8"?>` with `<?xml version="1.0" encoding="UTF-8" standalone="yes"?>`.
    # Support for `standalone` was added to `toprettyxml` in python 3.9, so do a manual work around.
    output[0] = b"""<?xml version="1.0" encoding="UTF-8" standalone="yes"?>"""

    output = b"\n".join(output)

    return output


def write_manifest_to_executable(filename, manifest_xml):
    """
    Write the given manifest XML to the given executable's RT_MANIFEST resource.
    """
    from PyInstaller.utils.win32 import winresource

    # CREATEPROCESS_MANIFEST_RESOURCE_ID is used for manifest resource in executables.
    # ISOLATIONAWARE_MANIFEST_RESOURCE_ID is used for manifest resources in DLLs.
    names = [CREATEPROCESS_MANIFEST_RESOURCE_ID]

    # Ensure LANG_NEUTRAL is updated, and also update any other present languages.
    languages = [LANG_NEUTRAL, "*"]

    winresource.add_or_update_resource(filename, manifest_xml, RT_MANIFEST, names, languages)


def read_manifest_from_executable(filename):
    """
    Read manifest from the given executable."
    """
    from PyInstaller.utils.win32 import winresource

    resources = winresource.get_resources(filename, [RT_MANIFEST])

    # `resources` is a three-level dictionary:
    #  - level 1: resource type (RT_MANIFEST)
    #  - level 2: resource name (CREATEPROCESS_MANIFEST_RESOURCE_ID)
    #  - level 3: resource language (LANG_NEUTRAL)

    # Level 1
    if RT_MANIFEST not in resources:
        raise ValueError(f"No RT_MANIFEST resources found in {filename!r}.")
    resources = resources[RT_MANIFEST]

    # Level 2
    if CREATEPROCESS_MANIFEST_RESOURCE_ID not in resources:
        raise ValueError(f"No RT_MANIFEST resource named CREATEPROCESS_MANIFEST_RESOURCE_ID found in {filename!r}.")
    resources = resources[CREATEPROCESS_MANIFEST_RESOURCE_ID]

    # Level 3
    # We prefer LANG_NEUTRAL, but allow fall back to the first available entry.
    if LANG_NEUTRAL in resources:
        resources = resources[LANG_NEUTRAL]
    else:
        resources = next(iter(resources.items()))

    manifest_xml = resources
    return manifest_xml



#

print('This is module ' + __name__)

# Tick: import a package in a way, PyInstaller is not able to track it. But we want to import it anyway to make the
# submodule `sub11` print it's name.
sub11 = ''.join('sub11')
__import__(__name__ + '.' + sub11, globals=globals(), level=0)



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

from PyInstaller.utils.tests import requires


@requires('six >= 1.0')
def test_six_moves(pyi_builder):
    pyi_builder.test_source("""
        from six.moves import UserList
        UserList
        """)


# Run the same test a second time to trigger errors like
#   Target module "six.moves.urllib" already imported as "AliasNode(…)"
# caused by PyiModuleGraph being cached in a insufficient way.
@requires('six >= 1.0')
def test_six_moves_2nd_run(pyi_builder):
    return test_six_moves(pyi_builder)



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import pytest

from PyInstaller.lib.modulegraph import modulegraph
from PyInstaller import configure
from PyInstaller import __main__ as pyi_main


@pytest.fixture
def large_import_chain(tmp_path):
    # Create package directory
    pkg = tmp_path / 'pkg'
    pkg.mkdir()
    # Create pkg/__init__py that imports first sub-package (a)
    (pkg / '__init__.py').write_text("from . import a", encoding='utf-8')
    # Create sub-packages
    mod = None
    for alpha in "abcdefg":
        # Overwrite last module of previous sub-package to import this sub-package.
        if mod:
            mod.write_text(f"import pkg.{alpha}", encoding='utf-8')
        # Create sub-package
        subpkg = pkg / alpha
        subpkg.mkdir()
        # Create sub-package's __init__.py, which imports first module.
        (subpkg / '__init__.py').write_text(f"from . import {alpha}000", encoding='utf-8')
        # Create modules; each module imports its next sibling (except the very last one, which we overwrite at the
        # start of next loop iteration).
        for num in range(250):
            mod = subpkg / f"{alpha}{num:03}.py"
            mod.write_text(f"from . import {alpha}{num + 1:03}", encoding='utf-8')

    script = tmp_path / 'script.py'
    script.write_text('import pkg', encoding='utf-8')

    return [str(tmp_path)], str(script)


def test_recursion_too_deep(large_import_chain):
    """
    modulegraph is recursive and triggers RecursionError if nesting of imported modules is too deep.
    This can be worked around by increasing recursion limit.

    With the default recursion limit (1000), the recursion error occurs at about 115 modules, with limit 2000
    (as tested below) at about 240 modules, and with limit 5000 at about 660 modules.
    """
    path, script = large_import_chain
    mg = modulegraph.ModuleGraph(path)
    # Increase recursion limit to 5 times of the default. Given the module import chain created above
    # this still should fail.
    with pytest.raises(RecursionError):
        mg.add_script(script)


def test_RecursionError_prints_message(tmp_path, large_import_chain, monkeypatch):
    """
    modulegraph is recursive and triggers RecursionError if nesting of imported modules is too deep.
    Ensure an informative message is printed if RecursionError occurs.
    """
    path, script = large_import_chain

    default_args = [
        '--specpath', str(tmp_path),
        '--distpath', str(tmp_path / 'dist'),
        '--workpath', str(tmp_path / 'build'),
        '--path', str(tmp_path),
    ]  # yapf: disable

    pyi_args = [script, *default_args]
    PYI_CONFIG = configure.get_config()
    PYI_CONFIG['cachedir'] = str(tmp_path)

    with pytest.raises(SystemExit) as execinfo:
        pyi_main.run(pyi_args, PYI_CONFIG)
    assert "sys.setrecursionlimit" in str(execinfo.value)



# ----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
# ----------------------------------------------------------------------------
"""
Various classes and functions to provide some backwards-compatibility with previous versions of Python onward.
"""
from __future__ import annotations

import errno

import importlib.machinery
import importlib.util
import os
import platform
import site
import subprocess
import sys
import sysconfig
import shutil
import types

from PyInstaller._shared_with_waf import _pyi_machine
from PyInstaller.exceptions import ExecCommandFailed

# hatch_build.py sets this environment variable to avoid errors due to unmet run-time dependencies. The
# PyInstaller.compat module is imported by hatch_build.py to build wheels, and some dependencies that are otherwise
# required at run-time (importlib-metadata on python < 3.10, pywin32-ctypes on Windows) might not be present while
# building wheels, nor are they required during that phase.
_setup_py_mode = os.environ.get('_PYINSTALLER_SETUP', '0') != '0'

# PyInstaller requires importlib.metadata from python >= 3.10 stdlib, or equivalent importlib-metadata >= 4.6.
if _setup_py_mode:
    importlib_metadata = None
else:
    if sys.version_info >= (3, 10):
        import importlib.metadata as importlib_metadata
    else:
        try:
            import importlib_metadata
        except ImportError as e:
            from PyInstaller.exceptions import ImportlibMetadataError
            raise ImportlibMetadataError() from e

        import packaging.version  # For importlib_metadata version check

        # Validate the version
        if packaging.version.parse(importlib_metadata.version("importlib-metadata")) < packaging.version.parse("4.6"):
            from PyInstaller.exceptions import ImportlibMetadataError
            raise ImportlibMetadataError()

# Strict collect mode, which raises error when trying to collect duplicate files into PKG/CArchive or COLLECT.
strict_collect_mode = os.environ.get("PYINSTALLER_STRICT_COLLECT_MODE", "0") != "0"

# Copied from https://docs.python.org/3/library/platform.html#cross-platform.
is_64bits: bool = sys.maxsize > 2**32

# Distinguish specific code for various Python versions. Variables 'is_pyXY' mean that Python X.Y and up is supported.
# Keep even unsupported versions here to keep 3rd-party hooks working.
is_py35 = sys.version_info >= (3, 5)
is_py36 = sys.version_info >= (3, 6)
is_py37 = sys.version_info >= (3, 7)
is_py38 = sys.version_info >= (3, 8)
is_py39 = sys.version_info >= (3, 9)
is_py310 = sys.version_info >= (3, 10)
is_py311 = sys.version_info >= (3, 11)
is_py312 = sys.version_info >= (3, 12)
is_py313 = sys.version_info >= (3, 13)
is_py314 = sys.version_info >= (3, 14)

is_win = sys.platform.startswith('win')
is_win_10 = is_win and (platform.win32_ver()[0] == '10')
is_win_11 = is_win and (platform.win32_ver()[0] == '11')
is_win_wine = False  # Running under Wine; determined later on.
is_cygwin = sys.platform == 'cygwin'
is_darwin = sys.platform == 'darwin'  # macOS

# Unix platforms
is_linux = sys.platform.startswith('linux')
is_solar = sys.platform.startswith('sun')  # Solaris
is_aix = sys.platform.startswith('aix')
is_freebsd = sys.platform.startswith('freebsd')
is_openbsd = sys.platform.startswith('openbsd')
is_hpux = sys.platform.startswith('hp-ux')

# Some code parts are similar to several unix platforms (e.g. Linux, Solaris, AIX).
# macOS is not considered as unix since there are many platform-specific details for Mac in PyInstaller.
is_unix = is_linux or is_solar or is_aix or is_freebsd or is_hpux or is_openbsd

# Linux distributions such as Alpine or OpenWRT use musl as their libc implementation and resultantly need specially
# compiled bootloaders. On musl systems, ldd with no arguments prints 'musl' and its version.
is_musl = is_linux and "musl" in subprocess.run(["ldd"], capture_output=True, encoding="utf-8").stderr

# Termux - terminal emulator and Linux environment app for Android.
is_termux = is_linux and hasattr(sys, 'getandroidapilevel')

# macOS version
_macos_ver = tuple(int(x) for x in platform.mac_ver()[0].split('.')) if is_darwin else None

# macOS 11 (Big Sur): if python is not compiled with Big Sur support, it ends up in compatibility mode by default, which
# is indicated by platform.mac_ver() returning '10.16'. The lack of proper Big Sur support breaks find_library()
# function from ctypes.util module, as starting with Big Sur, shared libraries are not visible on disk anymore. Support
# for the new library search mechanism was added in python 3.9 when compiled with Big Sur support. In such cases,
# platform.mac_ver() reports version as '11.x'. The behavior can be further modified via SYSTEM_VERSION_COMPAT
# environment variable; which allows explicitly enabling or disabling the compatibility mode. However, note that
# disabling the compatibility mode and using python that does not properly support Big Sur still leaves find_library()
# broken (which is a scenario that we ignore at the moment).
# The same logic applies to macOS 12 (Monterey).
is_macos_11_compat = bool(_macos_ver) and _macos_ver[0:2] == (10, 16)  # Big Sur or newer in compat mode
is_macos_11_native = bool(_macos_ver) and _macos_ver[0:2] >= (11, 0)  # Big Sur or newer in native mode
is_macos_11 = is_macos_11_compat or is_macos_11_native  # Big Sur or newer

# Check if python >= 3.13 was built with Py_GIL_DISABLED / free-threading (PEP703).
#
# This affects the shared library name, which has the "t" ABI suffix, as per:
# https://github.com/python/steering-council/issues/221#issuecomment-1841593283
#
# It also affects the layout of PyConfig structure used by bootloader; consequently we need to inform bootloader what
# kind of build it is dealing with (only in python 3.13; with 3.14 and later, we use PEP741 configuration API in the
# bootloader, and do not need to know the layout of PyConfig structure anymore)
is_nogil = bool(sysconfig.get_config_var('Py_GIL_DISABLED'))

# In a virtual environment created by virtualenv (github.com/pypa/virtualenv) there exists sys.real_prefix with the path
# to the base Python installation from which the virtual environment was created. This is true regardless of the version
# of Python used to execute the virtualenv command.
#
# In a virtual environment created by the venv module available in the Python standard lib, there exists sys.base_prefix
# with the path to the base implementation. This does not exist in a virtual environment created by virtualenv.
#
# The following code creates compat.is_venv and is.virtualenv that are True when running a virtual environment, and also
# compat.base_prefix with the path to the base Python installation.

base_prefix: str = os.path.abspath(getattr(sys, 'real_prefix', getattr(sys, 'base_prefix', sys.prefix)))
# Ensure `base_prefix` is not containing any relative parts.
is_venv = is_virtualenv = base_prefix != os.path.abspath(sys.prefix)

# Conda environments sometimes have different paths or apply patches to packages that can affect how a hook or package
# should access resources. Method for determining conda taken from https://stackoverflow.com/questions/47610844#47610844
is_conda = os.path.isdir(os.path.join(base_prefix, 'conda-meta'))

# Similar to ``is_conda`` but is ``False`` using another ``venv``-like manager on top. In this case, no packages
# encountered will be conda packages meaning that the default non-conda behaviour is generally desired from PyInstaller.
is_pure_conda = os.path.isdir(os.path.join(sys.prefix, 'conda-meta'))

# Full path to python interpreter.
python_executable = getattr(sys, '_base_executable', sys.executable)

# Is this Python from Microsoft App Store (Windows only)? Python from Microsoft App Store has executable pointing at
# empty shims.
is_ms_app_store = is_win and os.path.getsize(python_executable) == 0

if is_ms_app_store:
    # Locate the actual executable inside base_prefix.
    python_executable = os.path.join(base_prefix, os.path.basename(python_executable))
    if not os.path.exists(python_executable):
        raise SystemExit(
            'ERROR: PyInstaller cannot locate real python executable belonging to Python from Microsoft App Store!'
        )

# Bytecode magic value
BYTECODE_MAGIC = importlib.util.MAGIC_NUMBER

# List of suffixes for Python C extension modules.
EXTENSION_SUFFIXES = importlib.machinery.EXTENSION_SUFFIXES
ALL_SUFFIXES = importlib.machinery.all_suffixes()

# On Windows we require pywin32-ctypes.
# -> all pyinstaller modules should use win32api from PyInstaller.compat to
#    ensure that it can work on MSYS2 (which requires pywin32-ctypes)
if is_win:
    if _setup_py_mode:
        pywintypes = None
        win32api = None
    else:
        try:
            # Hide the `cffi` package from win32-ctypes by temporarily blocking its import. This ensures that `ctypes`
            # backend is always used, even if `cffi` is available. The `cffi` backend uses `pycparser`, which is
            # incompatible with -OO mode (2nd optimization level) due to its removal of docstrings.
            # See https://github.com/pyinstaller/pyinstaller/issues/6345
            # On the off chance that `cffi` has already been imported, store the `sys.modules` entry so we can restore
            # it after importing `pywin32-ctypes` modules.
            orig_cffi = sys.modules.get('cffi')
            sys.modules['cffi'] = None

            from win32ctypes.pywin32 import pywintypes  # noqa: F401, E402
            from win32ctypes.pywin32 import win32api  # noqa: F401, E402
        except ImportError as e:
            raise SystemExit(
                'ERROR: Could not import `pywintypes` or `win32api` from `win32ctypes.pywin32`.\n'
                'Please make sure that `pywin32-ctypes` is installed and importable, for example:\n\n'
                'pip install pywin32-ctypes\n'
            ) from e
        finally:
            # Unblock `cffi`.
            if orig_cffi is not None:
                sys.modules['cffi'] = orig_cffi
            else:
                del sys.modules['cffi']
            del orig_cffi

# macOS's platform.architecture() can be buggy, so we do this manually here. Based off the python documentation:
# https://docs.python.org/3/library/platform.html#platform.architecture
if is_darwin:
    architecture = '64bit' if sys.maxsize > 2**32 else '32bit'
else:
    architecture = platform.architecture()[0]

# Cygwin needs special handling, because platform.system() contains identifiers such as MSYS_NT-10.0-19042 and
# CYGWIN_NT-10.0-19042 that do not fit PyInstaller's OS naming scheme. Explicitly set `system` to 'Cygwin'.
system = 'Cygwin' if is_cygwin else platform.system()

# Machine suffix for bootloader.
if is_win:
    # On Windows ARM64 using an x64 Python environment, platform.machine() returns ARM64 but
    # we really want the bootloader that matches the Python environment instead of the OS.
    machine = _pyi_machine(os.environ.get("PROCESSOR_ARCHITECTURE", platform.machine()), platform.system())
else:
    machine = _pyi_machine(platform.machine(), platform.system())


# Wine detection and support
def is_wine_dll(filename: str | os.PathLike):
    """
    Check if the given PE file is a Wine DLL (PE-converted built-in, or fake/placeholder one).

    Returns True if the given file is a Wine DLL, False if not (or if file cannot be analyzed or does not exist).
    """
    _WINE_SIGNATURES = (
        b'Wine builtin DLL',  # PE-converted Wine DLL
        b'Wine placeholder DLL',  # Fake/placeholder Wine DLL
    )
    _MAX_LEN = max([len(sig) for sig in _WINE_SIGNATURES])

    # Wine places their DLL signature in the padding area between the IMAGE_DOS_HEADER and IMAGE_NT_HEADERS. So we need
    # to compare the bytes that come right after IMAGE_DOS_HEADER, i.e., after initial 64 bytes. We can read the file
    # directly and avoid using the pefile library to avoid performance penalty associated with full header parsing.
    try:
        with open(filename, 'rb') as fp:
            fp.seek(64)
            signature = fp.read(_MAX_LEN)
        return signature.startswith(_WINE_SIGNATURES)
    except Exception:
        pass
    return False


if is_win:
    try:
        import ctypes.util  # noqa: E402
        is_win_wine = is_wine_dll(ctypes.util.find_library('kernel32'))
    except Exception:
        pass

# Set and get environment variables does not handle unicode strings correctly on Windows.

# Acting on os.environ instead of using getenv()/setenv()/unsetenv(), as suggested in
# <http://docs.python.org/library/os.html#os.environ>: "Calling putenv() directly does not change os.environ, so it is
# better to modify os.environ." (Same for unsetenv.)


def getenv(name: str, default: str | None = None):
    """
    Returns unicode string containing value of environment variable 'name'.
    """
    return os.environ.get(name, default)


def setenv(name: str, value: str):
    """
    Accepts unicode string and set it as environment variable 'name' containing value 'value'.
    """
    os.environ[name] = value


def unsetenv(name: str):
    """
    Delete the environment variable 'name'.
    """
    # Some platforms (e.g., AIX) do not support `os.unsetenv()` and thus `del os.environ[name]` has no effect on the
    # real environment. For this case, we set the value to the empty string.
    os.environ[name] = ""
    del os.environ[name]


# Exec commands in subprocesses.


def exec_command(
    *cmdargs: str, encoding: str | None = None, raise_enoent: bool | None = None, **kwargs: int | bool | list | None
):
    """
    Run the command specified by the passed positional arguments, optionally configured by the passed keyword arguments.

    .. DANGER::
       **Ignore this function's return value** -- unless this command's standard output contains _only_ pathnames, in
       which case this function returns the correct filesystem-encoded string expected by PyInstaller. In all other
       cases, this function's return value is _not_ safely usable.

       For backward compatibility, this function's return value non-portably depends on the current Python version and
       passed keyword arguments:

       * Under Python 3.x, this value is a **decoded `str` string**. However, even this value is _not_ necessarily
         safely usable:
         * If the `encoding` parameter is passed, this value is guaranteed to be safely usable.
         * Else, this value _cannot_ be safely used for any purpose (e.g., string manipulation or parsing), except to be
           passed directly to another non-Python command. Why? Because this value has been decoded with the encoding
           specified by `sys.getfilesystemencoding()`, the encoding used by `os.fsencode()` and `os.fsdecode()` to
           convert from platform-agnostic to platform-specific pathnames. This is _not_ necessarily the encoding with
           which this command's standard output was encoded. Cue edge-case decoding exceptions.

    Parameters
    ----------
    cmdargs :
        Variadic list whose:
        1. Mandatory first element is the absolute path, relative path, or basename in the current `${PATH}` of the
           command to run.
        2. Optional remaining elements are arguments to pass to this command.
    encoding : str, optional
        Optional keyword argument specifying the encoding with which to decode this command's standard output under
        Python 3. As this function's return value should be ignored, this argument should _never_ be passed.
    raise_enoent : boolean, optional
        Optional keyword argument to simply raise the exception if the executing the command fails since to the command
        is not found. This is useful to checking id a command exists.

    All remaining keyword arguments are passed as is to the `subprocess.Popen()` constructor.

    Returns
    ----------
    str
        Ignore this value. See discussion above.
    """

    proc = subprocess.Popen(cmdargs, stdout=subprocess.PIPE, **kwargs)
    try:
        out = proc.communicate(timeout=60)[0]
    except OSError as e:
        if raise_enoent and e.errno == errno.ENOENT:
            raise
        print('--' * 20, file=sys.stderr)
        print("Error running '%s':" % " ".join(cmdargs), file=sys.stderr)
        print(e, file=sys.stderr)
        print('--' * 20, file=sys.stderr)
        raise ExecCommandFailed("ERROR: Executing command failed!") from e
    except subprocess.TimeoutExpired:
        proc.kill()
        raise

    # stdout/stderr are returned as a byte array NOT as string, so we need to convert that to proper encoding.
    try:
        if encoding:
            out = out.decode(encoding)
        else:
            # If no encoding is given, assume we are reading filenames from stdout only because it is the common case.
            out = os.fsdecode(out)
    except UnicodeDecodeError as e:
        # The sub-process used a different encoding; provide more information to ease debugging.
        print('--' * 20, file=sys.stderr)
        print(str(e), file=sys.stderr)
        print('These are the bytes around the offending byte:', file=sys.stderr)
        print('--' * 20, file=sys.stderr)
        raise
    return out


def exec_command_rc(*cmdargs: str, **kwargs: float | bool | list | None):
    """
    Return the exit code of the command specified by the passed positional arguments, optionally configured by the
    passed keyword arguments.

    Parameters
    ----------
    cmdargs : list
        Variadic list whose:
        1. Mandatory first element is the absolute path, relative path, or basename in the current `${PATH}` of the
           command to run.
        2. Optional remaining elements are arguments to pass to this command.

    All keyword arguments are passed as is to the `subprocess.call()` function.

    Returns
    ----------
    int
        This command's exit code as an unsigned byte in the range `[0, 255]`, where 0 signifies success and all other
        values signal a failure.
    """

    # 'encoding' keyword is not supported for 'subprocess.call'; remove it from kwargs.
    if 'encoding' in kwargs:
        kwargs.pop('encoding')
    return subprocess.call(cmdargs, **kwargs)


def exec_command_all(*cmdargs: str, encoding: str | None = None, **kwargs: int | bool | list | None):
    """
    Run the command specified by the passed positional arguments, optionally configured by the passed keyword arguments.

    .. DANGER::
       **Ignore this function's return value.** If this command's standard output consists solely of pathnames, consider
       calling `exec_command()`

    Parameters
    ----------
    cmdargs : str
        Variadic list whose:
        1. Mandatory first element is the absolute path, relative path, or basename in the current `${PATH}` of the
           command to run.
        2. Optional remaining elements are arguments to pass to this command.
    encoding : str, optional
        Optional keyword argument specifying the encoding with which to decode this command's standard output. As this
        function's return value should be ignored, this argument should _never_ be passed.

    All remaining keyword arguments are passed as is to the `subprocess.Popen()` constructor.

    Returns
    ----------
    (int, str, str)
        Ignore this 3-element tuple `(exit_code, stdout, stderr)`. See the `exec_command()` function for discussion.
    """
    proc = subprocess.Popen(
        cmdargs,
        bufsize=-1,  # Default OS buffer size.
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        **kwargs
    )
    # Waits for subprocess to complete.
    try:
        out, err = proc.communicate(timeout=60)
    except subprocess.TimeoutExpired:
        proc.kill()
        raise
    # stdout/stderr are returned as a byte array NOT as string. Thus we need to convert that to proper encoding.
    try:
        if encoding:
            out = out.decode(encoding)
            err = err.decode(encoding)
        else:
            # If no encoding is given, assume we're reading filenames from stdout only because it's the common case.
            out = os.fsdecode(out)
            err = os.fsdecode(err)
    except UnicodeDecodeError as e:
        # The sub-process used a different encoding, provide more information to ease debugging.
        print('--' * 20, file=sys.stderr)
        print(str(e), file=sys.stderr)
        print('These are the bytes around the offending byte:', file=sys.stderr)
        print('--' * 20, file=sys.stderr)
        raise

    return proc.returncode, out, err


def __wrap_python(args, kwargs):
    cmdargs = [sys.executable]

    # macOS supports universal binaries (binary for multiple architectures. We need to ensure that subprocess
    # binaries are running for the same architecture as python executable. It is necessary to run binaries with 'arch'
    # command.
    if is_darwin:
        if architecture == '64bit':
            if platform.machine() == 'arm64':
                py_prefix = ['arch', '-arm64']  # Apple M1
            else:
                py_prefix = ['arch', '-x86_64']  # Intel
        elif architecture == '32bit':
            py_prefix = ['arch', '-i386']
        else:
            py_prefix = []
        # Since macOS 10.11, the environment variable DYLD_LIBRARY_PATH is no more inherited by child processes, so we
        # proactively propagate the current value using the `-e` option of the `arch` command.
        if 'DYLD_LIBRARY_PATH' in os.environ:
            path = os.environ['DYLD_LIBRARY_PATH']
            py_prefix += ['-e', 'DYLD_LIBRARY_PATH=%s' % path]
        cmdargs = py_prefix + cmdargs

    if not __debug__:
        cmdargs.append('-O')

    cmdargs.extend(args)

    env = kwargs.get('env')
    if env is None:
        env = dict(**os.environ)

    # Ensure python 3 subprocess writes 'str' as utf-8
    env['PYTHONIOENCODING'] = 'UTF-8'
    # ... and ensure we read output as utf-8
    kwargs['encoding'] = 'UTF-8'

    return cmdargs, kwargs


def exec_python(*args: str, **kwargs: str | None):
    """
    Wrap running python script in a subprocess.

    Return stdout of the invoked command.
    """
    cmdargs, kwargs = __wrap_python(args, kwargs)
    return exec_command(*cmdargs, **kwargs)


def exec_python_rc(*args: str, **kwargs: str | None):
    """
    Wrap running python script in a subprocess.

    Return exit code of the invoked command.
    """
    cmdargs, kwargs = __wrap_python(args, kwargs)
    return exec_command_rc(*cmdargs, **kwargs)


# Path handling.


# Site-packages functions - use native function if available.
def getsitepackages(prefixes: list | None = None):
    """
    Returns a list containing all global site-packages directories.

    For each directory present in ``prefixes`` (or the global ``PREFIXES``), this function finds its `site-packages`
    subdirectory depending on the system environment, and returns a list of full paths.
    """
    # This implementation was copied from the ``site`` module, python 3.7.3.
    sitepackages = []
    seen = set()

    if prefixes is None:
        prefixes = [sys.prefix, sys.exec_prefix]

    for prefix in prefixes:
        if not prefix or prefix in seen:
            continue
        seen.add(prefix)

        if os.sep == '/':
            sitepackages.append(os.path.join(prefix, "lib", "python%d.%d" % sys.version_info[:2], "site-packages"))
        else:
            sitepackages.append(prefix)
            sitepackages.append(os.path.join(prefix, "lib", "site-packages"))
    return sitepackages


# Backported for virtualenv. Module 'site' in virtualenv might not have this attribute.
getsitepackages = getattr(site, 'getsitepackages', getsitepackages)


# Wrapper to load a module from a Python source file. This function loads import hooks when processing them.
def importlib_load_source(name: str, pathname: str):
    # Import module from a file.
    mod_loader = importlib.machinery.SourceFileLoader(name, pathname)
    mod = types.ModuleType(mod_loader.name)
    mod.__file__ = mod_loader.get_filename()  # Some hooks require __file__ attribute in their namespace
    mod_loader.exec_module(mod)
    return mod


# Patterns of module names that should be bundled into the base_library.zip to be available during bootstrap.
# These modules include direct or indirect dependencies of encodings.* modules. The encodings modules must be
# recursively included to set the I/O encoding during python startup. Similarly, this list should include
# modules used by PyInstaller's bootstrap scripts and modules (loader/pyi*.py)

PY3_BASE_MODULES = {
    '_collections_abc',
    '_weakrefset',
    'abc',
    'codecs',
    'collections',
    'copyreg',
    'encodings',
    'enum',
    'functools',
    'genericpath',  # dependency of os.path
    'io',
    'heapq',
    'keyword',
    'linecache',
    'locale',
    'ntpath',  # dependency of os.path
    'operator',
    'os',
    'posixpath',  # dependency of os.path
    're',
    'reprlib',
    'sre_compile',
    'sre_constants',
    'sre_parse',
    'stat',  # dependency of os.path
    'traceback',  # for startup errors
    'types',
    'weakref',
    'warnings',
}

if not is_py310:
    PY3_BASE_MODULES.add('_bootlocale')

# Object types of Pure Python modules in modulegraph dependency graph.
# Pure Python modules have code object (attribute co_code).
PURE_PYTHON_MODULE_TYPES = {
    'SourceModule',
    'CompiledModule',
    'Package',
    'NamespacePackage',
    # Deprecated.
    # TODO Could these module types be removed?
    'FlatPackage',
    'ArchiveModule',
}
# Object types of special Python modules (built-in, run-time, namespace package) in modulegraph dependency graph that do
# not have code object.
SPECIAL_MODULE_TYPES = {
    # Omit AliasNode from here (and consequently from VALID_MODULE_TYPES), in order to prevent PyiModuleGraph from
    # running standard hooks for aliased modules.
    #'AliasNode',
    'BuiltinModule',
    'RuntimeModule',
    'RuntimePackage',

    # PyInstaller handles scripts differently and not as standard Python modules.
    'Script',
}
# Object types of Binary Python modules (extensions, etc) in modulegraph dependency graph.
BINARY_MODULE_TYPES = {
    'Extension',
    'ExtensionPackage',
}
# Object types of valid Python modules in modulegraph dependency graph.
VALID_MODULE_TYPES = PURE_PYTHON_MODULE_TYPES | SPECIAL_MODULE_TYPES | BINARY_MODULE_TYPES
# Object types of bad/missing/invalid Python modules in modulegraph dependency graph.
# TODO: should be 'Invalid' module types also in the 'MISSING' set?
BAD_MODULE_TYPES = {
    'BadModule',
    'ExcludedModule',
    'InvalidSourceModule',
    'InvalidCompiledModule',
    'MissingModule',

    # Runtime modules and packages are technically valid rather than bad, but exist only in-memory rather than on-disk
    # (typically due to pre_safe_import_module() hooks), and hence cannot be physically frozen. For simplicity, these
    # nodes are categorized as bad rather than valid.
    'RuntimeModule',
    'RuntimePackage',
}
ALL_MODULE_TYPES = VALID_MODULE_TYPES | BAD_MODULE_TYPES
# TODO: review this mapping to TOC, remove useless entries.
# Dictionary to map ModuleGraph node types to TOC typecodes.
MODULE_TYPES_TO_TOC_DICT = {
    # Pure modules.
    'AliasNode': 'PYMODULE',
    'Script': 'PYSOURCE',
    'SourceModule': 'PYMODULE',
    'CompiledModule': 'PYMODULE',
    'Package': 'PYMODULE',
    'FlatPackage': 'PYMODULE',
    'ArchiveModule': 'PYMODULE',
    # Binary modules.
    'Extension': 'EXTENSION',
    'ExtensionPackage': 'EXTENSION',
    # Special valid modules.
    'BuiltinModule': 'BUILTIN',
    'NamespacePackage': 'PYMODULE',
    # Bad modules.
    'BadModule': 'bad',
    'ExcludedModule': 'excluded',
    'InvalidSourceModule': 'invalid',
    'InvalidCompiledModule': 'invalid',
    'MissingModule': 'missing',
    'RuntimeModule': 'runtime',
    'RuntimePackage': 'runtime',
    # Other.
    'does not occur': 'BINARY',
}


def check_requirements():
    """
    Verify that all requirements to run PyInstaller are met.

    Fail hard if any requirement is not met.
    """
    # Fail hard if Python does not have minimum required version
    if sys.version_info < (3, 8):
        raise EnvironmentError('PyInstaller requires Python 3.8 or newer.')

    if sys.implementation.name != "cpython":
        raise SystemExit(f"ERROR: PyInstaller does not support {sys.implementation.name}. Only CPython is supported.")

    if getattr(sys, "frozen", False):
        raise SystemExit("ERROR: PyInstaller can not be ran on itself")

    # There are some old packages which used to be backports of libraries which are now part of the standard library.
    # These backports are now unmaintained and contain only an older subset of features leading to obscure errors like
    # "enum has not attribute IntFlag" if installed.
    from importlib.metadata import distribution, PackageNotFoundError

    for name in ["enum34", "typing", "pathlib"]:
        try:
            dist = distribution(name)
        except PackageNotFoundError:
            continue
        remove = "conda remove" if is_conda else f'"{sys.executable}" -m pip uninstall {name}'
        raise SystemExit(
            f"ERROR: The '{name}' package is an obsolete backport of a standard library package and is incompatible "
            f"with PyInstaller. Please remove this package (located in {dist.locate_file('')}) using\n    {remove}\n"
            "then try again."
        )

    # Bail out if binutils is not installed.
    if is_linux and shutil.which("objdump") is None:
        raise SystemExit(
            "ERROR: On Linux, objdump is required. It is typically provided by the 'binutils' package "
            "installable via your Linux distribution's package manager."
        )



# ----------------------------------------------------------------------------
# Copyright (c) 2024, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
from PyInstaller import log as logging
from PyInstaller import isolated

logger = logging.getLogger(__name__)


# Import setuptools and analyze its properties in an isolated subprocess. This function is called by `SetuptoolsInfo`
# to initialize its properties.
@isolated.decorate
def _retrieve_setuptools_info():
    import importlib

    try:
        setuptools = importlib.import_module("setuptools")  # noqa: F841
    except ModuleNotFoundError:
        return None

    # Delay these imports until after we have confirmed that setuptools is importable.
    import pathlib

    import packaging.version

    from PyInstaller.compat import importlib_metadata
    from PyInstaller.utils.hooks import (
        collect_data_files,
        collect_submodules,
    )

    # Try to retrieve the version. At this point, failure is consider an error.
    version_string = importlib_metadata.version("setuptools")
    version = packaging.version.Version(version_string).release  # Use the version tuple

    # setuptools >= 60.0 its vendored copy of distutils (mainly due to its removal from stdlib in python >= 3.12).
    distutils_vendored = False
    distutils_modules = []
    if version >= (60, 0):
        distutils_vendored = True
        distutils_modules += ["_distutils_hack"]
        distutils_modules += collect_submodules(
            "setuptools._distutils",
            # setuptools 71.0.1 ~ 71.0.4 include `setuptools._distutils.tests`; avoid explicitly collecting it
            # (t was not included in earlier setuptools releases).
            filter=lambda name: name != 'setuptools._distutils.tests',
        )

    # Check if `setuptools._vendor` exists. Some linux distributions opt to de-vendor `setuptools` and remove the
    # `setuptools._vendor` directory altogether. If this is the case, most of additional processing below should be
    # skipped to avoid errors and warnings about non-existent `setuptools._vendor` module.
    try:
        setuptools_vendor = importlib.import_module("setuptools._vendor")
    except ModuleNotFoundError:
        setuptools_vendor = None

    # Check for exposed packages/modules that are vendored by setuptools. If stand-alone version is not provided in the
    # environment, setuptools-vendored version is exposed (due to location of `setuptools._vendor` being appended to
    # `sys.path`. Applicable to v71.0.0 and later.
    vendored_status = dict()
    vendored_namespace_package_paths = dict()
    if version >= (71, 0) and setuptools_vendor is not None:
        VENDORED_TOP_LEVEL_NAMESPACE_CANDIDATES = (
            "backports",  # "regular" package, but has namespace semantics due to `pkgutil.extend_path()`
            "jaraco",  # PEP-420 namespace package
        )

        VENDORED_CANDIDATES = (
            "autocommand",
            "backports.tarfile",
            "importlib_metadata",
            "importlib_resources",
            "inflect",
            "jaraco.context",
            "jaraco.functools",
            "jaraco.text",
            "more_itertools",
            "ordered_set",
            "packaging",
            "platformdirs",
            "tomli",
            "typeguard",
            "typing_extensions",
            "wheel",
            "zipp",
        )

        # Resolve path(s) of `setuptools_vendor` package.
        setuptools_vendor_paths = [pathlib.Path(path).resolve() for path in setuptools_vendor.__path__]

        # Process each candidate: top-level namespace packages
        for candidate_name in VENDORED_TOP_LEVEL_NAMESPACE_CANDIDATES:
            try:
                candidate = importlib.import_module(candidate_name)
            except ImportError:
                continue

            # Retrieve the __path__ attribute and store it, so we can re-use it in hooks without having to re-import
            # `setuptools` and the candidate package...
            candidate_path_attr = getattr(candidate, '__path__', [])
            if candidate_path_attr:
                candidate_paths = [pathlib.Path(path).resolve() for path in candidate_path_attr]
                is_vendored = [
                    any([
                        setuptools_vendor_path in candidate_path.parents or candidate_path == setuptools_vendor_path
                        for setuptools_vendor_path in setuptools_vendor_paths
                    ]) for candidate_path in candidate_paths
                ]
                # For namespace packages, distinguish between "fully" vendored and "partially" vendored state; i.e.,
                # whether the part of namespace package in the vendored directory is the only part or not.
                if all(is_vendored):
                    vendored_status[candidate_name] = 'fully'
                elif any(is_vendored):
                    vendored_status[candidate_name] = 'partially'
                else:
                    vendored_status[candidate_name] = False

            # Store paths
            vendored_namespace_package_paths[candidate_name] = [str(path) for path in candidate_path_attr]

        # Process each candidate: modules and packages
        for candidate_name in VENDORED_CANDIDATES:
            try:
                candidate = importlib.import_module(candidate_name)
            except ImportError:
                continue

            # Check the __file__ attribute (modules and regular packages). Will not work with namespace packages, but
            # at the moment, there are none (vendored top-level namespace packages have already been handled).
            candidate_file_attr = getattr(candidate, '__file__', None)
            if candidate_file_attr is not None:
                candidate_path = pathlib.Path(candidate_file_attr).parent.resolve()
                is_vendored = any([
                    setuptools_vendor_path in candidate_path.parents or candidate_path == setuptools_vendor_path
                    for setuptools_vendor_path in setuptools_vendor_paths
                ])
                vendored_status[candidate_name] = is_vendored  # True/False

    # Collect submodules from `setuptools._vendor`, regardless of whether the vendored package is exposed or
    # not (because setuptools might need/use it either way).
    vendored_modules = []
    if setuptools_vendor is not None:
        EXCLUDED_VENDORED_MODULES = (
            # Prevent recursing into setuptools._vendor.pyparsing.diagram, which typically fails to be imported due
            # to missing dependencies (railroad, pyparsing (?), jinja2) and generates a warning... As the module is
            # usually unimportable, it is likely not to be used by setuptools. NOTE: pyparsing was removed from
            # vendored packages in setuptools v67.0.0; keep this exclude around for earlier versions.
            'setuptools._vendor.pyparsing.diagram',
            # Setuptools >= 71 started shipping vendored dependencies that include tests; avoid collecting those via
            # hidden imports. (Note that this also prevents creation of aliases for these module, but that should
            # not be an issue, as they should not be referenced from anywhere).
            'setuptools._vendor.importlib_resources.tests',
            # These appear to be utility scripts bundled with the jaraco.text package - exclude them.
            'setuptools._vendor.jaraco.text.show-newlines',
            'setuptools._vendor.jaraco.text.strip-prefix',
            'setuptools._vendor.jaraco.text.to-dvorak',
            'setuptools._vendor.jaraco.text.to-qwerty',
        )
        vendored_modules += collect_submodules(
            'setuptools._vendor',
            filter=lambda name: name not in EXCLUDED_VENDORED_MODULES,
        )

        # `collect_submodules` (and its underlying `pkgutil.iter_modules` do not discover namespace sub-packages, in
        # this case `setuptools._vendor.jaraco`. So force a manual scan of modules/packages inside it.
        vendored_modules += collect_submodules(
            'setuptools._vendor.jaraco',
            filter=lambda name: name not in EXCLUDED_VENDORED_MODULES,
        )

    # *** Data files for vendored packages ***
    vendored_data = []

    if version >= (71, 0) and setuptools_vendor is not None:
        # Since the vendored dependencies from `setuptools/_vendor` are now visible to the outside world, make
        # sure we collect their metadata. (We cannot use copy_metadata here, because we need to collect data
        # files to their original locations).
        vendored_data += collect_data_files('setuptools._vendor', includes=['**/*.dist-info'])
        # Similarly, ensure that `Lorem ipsum.txt` from vendored jaraco.text is collected
        vendored_data += collect_data_files('setuptools._vendor.jaraco.text', includes=['**/Lorem ipsum.txt'])

    # Return dictionary with collected information
    return {
        "available": True,
        "version": version,
        "distutils_vendored": distutils_vendored,
        "distutils_modules": distutils_modules,
        "vendored_status": vendored_status,
        "vendored_modules": vendored_modules,
        "vendored_data": vendored_data,
        "vendored_namespace_package_paths": vendored_namespace_package_paths,
    }


class SetuptoolsInfo:
    def __init__(self):
        pass

    def __repr__(self):
        return "SetuptoolsInfo"

    # Delay initialization of setuptools information until until the corresponding attributes are first requested.
    def __getattr__(self, name):
        if 'available' in self.__dict__:
            # Initialization was already done, but requested attribute is not available.
            raise AttributeError(name)

        # Load setuptools info...
        self._load_setuptools_info()
        # ... and return the requested attribute
        return getattr(self, name)

    def _load_setuptools_info(self):
        logger.info("%s: initializing cached setuptools info...", self)

        # Initialize variables so that they might be accessed even if setuptools is unavailable or if initialization
        # fails for some reason.
        self.available = False
        self.version = None
        self.distutils_vendored = False
        self.distutils_modules = []
        self.vendored_status = dict()
        self.vendored_modules = []
        self.vendored_data = []
        self.vendored_namespace_package_paths = dict()

        try:
            setuptools_info = _retrieve_setuptools_info()
        except Exception as e:
            logger.warning("%s: failed to obtain setuptools info: %s", self, e)
            return

        # If package could not be imported, `_retrieve_setuptools_info` returns None. In such cases, emit a debug
        # message instead of a warning, because this initialization might be triggered by a helper function that is
        # trying to determine availability of `setuptools` by inspecting the `available` attribute.
        if setuptools_info is None:
            logger.debug("%s: failed to obtain setuptools info: setuptools could not be imported.", self)
            return

        # Copy properties
        for key, value in setuptools_info.items():
            setattr(self, key, value)

    def is_vendored(self, module_name):
        return self.vendored_status.get(module_name, False)

    @staticmethod
    def _create_vendored_aliases(vendored_name, module_name, modules_list):
        # Create aliases for all submodules
        prefix_len = len(vendored_name)  # Length of target-name prefix to remove
        return ((module_name + vendored_module[prefix_len:], vendored_module) for vendored_module in modules_list
                if vendored_module.startswith(vendored_name))

    def get_vendored_aliases(self, module_name):
        vendored_name = f"setuptools._vendor.{module_name}"
        return self._create_vendored_aliases(vendored_name, module_name, self.vendored_modules)

    def get_distutils_aliases(self):
        vendored_name = "setuptools._distutils"
        return self._create_vendored_aliases(vendored_name, "distutils", self.distutils_modules)


setuptools_info = SetuptoolsInfo()


def pre_safe_import_module_for_top_level_namespace_packages(api):
    """
    A common implementation of pre_safe_import_module hook function for handling vendored top-level namespace packages
    (i.e., `backports` and `jaraco`).

    This function can be either called from the `pre_safe_import_module` function in a pre-safe-import-module hook, or
    just imported into the hook and aliased to `pre_safe_import_module`.
    """
    module_name = api.module_name

    # Check if the package/module is a vendored copy. This also returns False is setuptools is unavailable, because
    # vendored module status dictionary will be empty.
    vendored = setuptools_info.is_vendored(module_name)
    if not vendored:
        return

    if vendored == 'fully':
        # For a fully-vendored copy, force creation of aliases; on one hand, this aims to ensure that submodules are
        # resolvable, but on the other, it also prevents creation of unvendored top-level package, which should not
        # exit in this case.
        vendored_name = f"setuptools._vendor.{module_name}"
        logger.info(
            "Setuptools: %r appears to be a full setuptools-vendored copy - creating alias to %r!", module_name,
            vendored_name
        )
        # Create aliases for all (sub)modules
        for aliased_name, real_vendored_name in setuptools_info.get_vendored_aliases(module_name):
            api.add_alias_module(real_vendored_name, aliased_name)
    elif vendored == 'partially':
        # For a partially-vendored copy, adjust the submodule search paths, so that submodules from all locations are
        # discoverable (especially from the setuptools vendor directory, which might not be in the search path yet).
        search_paths = setuptools_info.vendored_namespace_package_paths.get(module_name, [])
        logger.info(
            "Setuptools: %r appears to be a partial setuptools-vendored copy - extending search paths to %r!",
            module_name, search_paths
        )
        for path in search_paths:
            api.append_package_path(path)
    else:
        logger.warning("Setuptools: %r has unhandled vendored status: %r", module_name, vendored)


def pre_safe_import_module(api):
    """
    A common implementation of pre_safe_import_module hook function.

    This function can be either called from the `pre_safe_import_module` function in a pre-safe-import-module hook, or
    just imported into the hook.
    """
    module_name = api.module_name

    # Check if the package/module is a vendored copy. This also returns False is setuptools is unavailable, because
    # vendored module status dictionary will be empty.
    if not setuptools_info.is_vendored(module_name):
        return

    vendored_name = f"setuptools._vendor.{module_name}"
    logger.info(
        "Setuptools: %r appears to be a setuptools-vendored copy - creating alias to %r!", module_name, vendored_name
    )

    # Create aliases for all (sub)modules
    for aliased_name, real_vendored_name in setuptools_info.get_vendored_aliases(module_name):
        api.add_alias_module(real_vendored_name, aliased_name)



# Different forms of relative-import for an optional sub-module. While the said sub-module does not exist, an
# unrelated eponymous top-level module does - and these import attempts should *NOT* trigger its collection!
try:
    from . import myotherpackage  # noqa: F401
except ImportError:
    pass

try:
    from .myotherpackage import something  # noqa: F401
except ImportError:
    pass



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# In dist directory are Python C-extension file names, e.g., module.submodule.so
# For example: ./simplejson/_speedups.so -> ./simplejson._speedups.so

import os
import sys

# List of suffixes for Python C extension modules.
from importlib.machinery import EXTENSION_SUFFIXES

from simplejson import _speedups

modpath = os.path.join(sys.prefix, 'simplejson._speedups')
frozen_modpath = _speedups.__file__

print('Module path expected:', modpath, '+ ext', file=sys.stderr)
print('Module path  current:', frozen_modpath, file=sys.stderr)

# Filename extensions can include several dots (e.g., '.cpython-33m.so'), so we cannot simply use os.path.splitext(),
# but have to loop over all possible extensions.
for ext in EXTENSION_SUFFIXES:
    if modpath + ext == frozen_modpath:
        print('             matched:', modpath + ext, file=sys.stderr)
        break
else:
    if not getattr(sys, 'frozen', False):
        raise SystemExit('This script only works correctly when frozen')
    raise SystemExit('Python C-extension file name is not correct.')



# fake module to make PyQt5 hooks and run-time-hooks happy
__pyinstaller_fake_module_marker__ = '__pyinstaller_fake_module_marker__'


class QLibraryInfo:
    def __init__(*args, **kw):
        pass

    PrefixPath = 1
    BinariesPath = 2

    @classmethod
    def location(cls, val):
        return "."

    @classmethod
    def isDebugBuild(cls):
        return False



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
Mock module of the same name as but _not_ shadowed by a global variable defined by the `__init__`
submodule of this package.

This module is exercised by the `test_import_submodule_global_unshadowed` functional test.
"""



#!/usr/bin/env python3
#-----------------------------------------------------------------------------
# Copyright (c) 2015-2023, PyInstaller Development Team.
# Copyright (c) 2015-2020, Hartmut Goebel.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
This script reformats the output of `myprog --help` to decent rst.

There are two sphinx plugins which could replace this eventually:
  - https://github.com/ashb/sphinx-argparse
  - https://github.com/gaborbernat/sphinx-argparse-cli
The former's output looks really nice, but lacks support for cross-referencing arguments. The latter supports
cross-referencing, but its output looks hideous. Hopefully either one of them will eventually evolve into something
we can use.

The main functions here are parser_to_rst() and help_to_rst(). Throughout this code the, **cross_references** option
controls whether or not section headings and options should be given cross reference targets.
"""

from textwrap import indent, wrap, dedent
import re
from argparse import ArgumentParser
from functools import partial


def parser_to_rst(parser: ArgumentParser, cross_references=True):
    """
    Extract the ``--help`` output from an argparse parser and convert it to restructured text.
    """
    help = parser.format_help()
    return help_to_rst(help, cross_references)


# Matches headings followed by indented blocks.
SECTION_REGEX = re.compile(
    r"""
    # A non-empty line with no indentation.
    ^\S.*\n

    # Followed by a non-zero number of either blank lines or indented lines.
    (?:(?:\ +.*)?\n)+
""", re.MULTILINE | re.VERBOSE
)


def help_to_rst(help: str, cross_references=True):
    """
    Convert the output of a ``cli --help`` call to rst.
    """
    summary, *sections = SECTION_REGEX.findall(help)

    # We could stick the summary (``usage: pyinstaller [-h] [--help] ...``) into a code block, but it is pretty
    # unhelpful, so I choose to omit it.
    sections = "\n".join(section_to_rst(section, cross_references) for section in sections)

    return sections


def section_to_rst(section: str, cross_references=True) -> str:
    """
    Convert a single option group's ``--help`` output to rst.

    This generates a heading for the option group followed by each option within that group.
    """
    title, body = section.split("\n", maxsplit=1)

    rst_title = rst_headerise(title, cross_references)
    rst_body = OPTION_REGEX.sub(partial(option_to_rst, cross_references=cross_references), body)

    return rst_title + rst_body


OPTION_REGEX = re.compile(
    r"""
    # Matches:
    #   --name, --other-name, -n VALUE  Some description
    #                                   and some more description.

    # An option name prefixed with at least 1 space.
    ^(\ +)(.*?)
    # Optionally followed by at least 2 spaces and the start of the description.
    (?:\ {2,}(.*))?\n

    # More lines of description.
    # Each line starts with more spaces than which prefixed the option name (so as to avoid picking up the next option).
    # Blank lines are allowed.
    (((?:\1\ +(.*))?\n)*)

""", re.MULTILINE | re.VERBOSE
)


def option_to_rst(m: re.Match, cross_references=True) -> str:
    """
    Convert a single option to rst.

    The output should look like::

        .. option:: --option-name -n

            The help for that option nicely text-wrapped.
    """
    name = m.group(2)
    assert name
    body = " ".join(i for i in m.group(3, 4) if i)
    # Escape characters which turn into invalid rst.
    body = body.replace("*", r"\*")
    # Re-wrap the help block.
    body = "\n".join(wrap(dedent(body), width=75, break_on_hyphens=False, break_long_words=False))

    template = ".. option:: {}\n\n{}\n\n" if cross_references else "{}\n\n{}\n\n"

    return template.format(name, indent(body, "    "))


def rst_headerise(title: str, cross_references=True) -> str:
    """
    Create a title with the correct length '---' underline.
    """
    title = title.strip(" \n:").title()
    out = f"{title}\n{'-' * len(title)}\n\n"
    if cross_references:
        out = f".. _`{title}`:\n\n" + out
    return out



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import os, re
from waflib import Utils, Options, Context

gnuopts = '''
bindir, user commands, ${EXEC_PREFIX}/bin
sbindir, system binaries, ${EXEC_PREFIX}/sbin
libexecdir, program-specific binaries, ${EXEC_PREFIX}/libexec
sysconfdir, host-specific configuration, ${PREFIX}/etc
sharedstatedir, architecture-independent variable data, ${PREFIX}/com
localstatedir, variable data, ${PREFIX}/var
libdir, object code libraries, ${EXEC_PREFIX}/lib%s
includedir, header files, ${PREFIX}/include
oldincludedir, header files for non-GCC compilers, /usr/include
datarootdir, architecture-independent data root, ${PREFIX}/share
datadir, architecture-independent data, ${DATAROOTDIR}
infodir, GNU "info" documentation, ${DATAROOTDIR}/info
localedir, locale-dependent data, ${DATAROOTDIR}/locale
mandir, manual pages, ${DATAROOTDIR}/man
docdir, documentation root, ${DATAROOTDIR}/doc/${PACKAGE}
htmldir, HTML documentation, ${DOCDIR}
dvidir, DVI documentation, ${DOCDIR}
pdfdir, PDF documentation, ${DOCDIR}
psdir, PostScript documentation, ${DOCDIR}
''' % Utils.lib64()
_options = [x.split(', ') for x in gnuopts.splitlines() if x]


def configure(conf):
    def get_param(varname, default):
        return getattr(Options.options, varname, '') or default

    env = conf.env
    env.LIBDIR = env.BINDIR = []
    env.EXEC_PREFIX = get_param('EXEC_PREFIX', env.PREFIX)
    env.PACKAGE = getattr(Context.g_module, 'APPNAME', None) or env.PACKAGE
    complete = False
    iter = 0
    while not complete and iter < len(_options) + 1:
        iter += 1
        complete = True
        for name, help, default in _options:
            name = name.upper()
            if not env[name]:
                try:
                    env[name] = Utils.subst_vars(get_param(name, default).replace('/', os.sep), env)
                except TypeError:
                    complete = False
    if not complete:
        lst = [x for x, _, _ in _options if not env[x.upper()]]
        raise conf.errors.WafError('Variable substitution failure %r' % lst)


def options(opt):
    inst_dir = opt.add_option_group(
        'Installation prefix', 'By default, "waf install" will put the files in\
 "/usr/local/bin", "/usr/local/lib" etc. An installation prefix other\
 than "/usr/local" can be given using "--prefix", for example "--prefix=$HOME"'
    )
    for k in ('--prefix', '--destdir'):
        option = opt.parser.get_option(k)
        if option:
            opt.parser.remove_option(k)
            inst_dir.add_option(option)
    inst_dir.add_option(
        '--exec-prefix', help='installation prefix for binaries [PREFIX]', default='', dest='EXEC_PREFIX'
    )
    dirs_options = opt.add_option_group('Installation directories')
    for name, help, default in _options:
        option_name = '--' + name
        str_default = default
        str_help = '%s [%s]' % (help, re.sub(r'\$\{([^}]+)\}', r'\1', str_default))
        dirs_options.add_option(option_name, help=str_help, default='', dest=name.upper())



#-----------------------------------------------------------------------------
# Copyright (c) 2021-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

from PyInstaller.utils.hooks import check_requirement

# As of scipy 1.6.0, scipy.spatial.transform.rotation is cython-compiled, so we fail to automatically pick up its
# imports.
if check_requirement("scipy >= 1.6.0"):
    hiddenimports = ['scipy.spatial.transform._rotation_groups']



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import re
import importlib.util

from PyInstaller import isolated
from PyInstaller.lib.modulegraph.modulegraph import SourceModule
from PyInstaller.utils.hooks import check_requirement, collect_entry_point, logger

datas = []

# 'sqlalchemy.testing' causes bundling a lot of unnecessary modules.
excludedimports = ['sqlalchemy.testing']

# Include most common database bindings some database bindings are detected and include some are not. We should
# explicitly include database backends.
hiddenimports = ['pysqlite2', 'MySQLdb', 'psycopg2', 'sqlalchemy.ext.baked']

if check_requirement('sqlalchemy >= 1.4'):
    hiddenimports.append("sqlalchemy.sql.default_comparator")


@isolated.decorate
def _get_dialect_modules(module_name):
    import importlib
    module = importlib.import_module(module_name)
    return [f"{module_name}.{submodule_name}" for submodule_name in module.__all__]


# In SQLAlchemy >= 0.6, the "sqlalchemy.dialects" package provides dialects.
# In SQLAlchemy <= 0.5, the "sqlalchemy.databases" package provides dialects.
if check_requirement('sqlalchemy >= 0.6'):
    hiddenimports += _get_dialect_modules("sqlalchemy.dialects")
else:
    hiddenimports += _get_dialect_modules("sqlalchemy.databases")

# Collect additional dialects and plugins that are registered via entry-points, under assumption that they are available
# in the build environment for a reason (i.e., they are used).
for entry_point_name in ('sqlalchemy.dialects', 'sqlalchemy.plugins'):
    ep_datas, ep_hiddenimports = collect_entry_point(entry_point_name)
    datas += ep_datas
    hiddenimports += ep_hiddenimports


def hook(hook_api):
    """
    SQLAlchemy 0.9 introduced the decorator 'util.dependencies'.  This decorator does imports. E.g.:

            @util.dependencies("sqlalchemy.sql.schema")

    This hook scans for included SQLAlchemy modules and then scans those modules for any util.dependencies and marks
    those modules as hidden imports.
    """

    if not check_requirement('sqlalchemy >= 0.9'):
        return

    # this parser is very simplistic but seems to catch all cases as of V1.1
    depend_regex = re.compile(r'@util.dependencies\([\'"](.*?)[\'"]\)')

    hidden_imports_set = set()
    known_imports = set()
    for node in hook_api.module_graph.iter_graph(start=hook_api.module):
        if isinstance(node, SourceModule) and node.identifier.startswith('sqlalchemy.'):
            known_imports.add(node.identifier)

            # Read the source...
            with open(node.filename, 'rb') as f:
                source_code = f.read()
            source_code = importlib.util.decode_source(source_code)

            # ... and scan it
            for match in depend_regex.findall(source_code):
                hidden_imports_set.add(match)

    hidden_imports_set -= known_imports
    if len(hidden_imports_set):
        logger.info("  Found %d sqlalchemy hidden imports", len(hidden_imports_set))
        hook_api.add_imports(*list(hidden_imports_set))



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import pathlib
import importlib.machinery

from PyInstaller import compat
from PyInstaller.depend import analysis, bindepend
from PyInstaller.building.build_main import Analysis
from PyInstaller.building.api import PYZ

_MODULES_DIR = pathlib.Path(__file__).parent / "modules"


def test_issue_2492(monkeypatch, tmp_path):
    # Crash if an extension module has an hidden import to ctypes (e.g. added by the hook).

    # Need to set up some values
    monkeypatch.setattr(
        'PyInstaller.config.CONF', {
            'workpath': str(tmp_path),
            'spec': str(tmp_path),
            'warnfile': str(tmp_path / 'warn.txt'),
            'dot-file': str(tmp_path / 'imports.dot'),
            'xref-file': str(tmp_path / 'imports.xref'),
            'hiddenimports': [],
            'specnm': 'issue_2492_script',
            'code_cache': dict(),
        }
    )
    # Speedup: avoid analyzing base_library.zip
    monkeypatch.setattr(analysis, 'PY3_BASE_MODULES', [])

    script = tmp_path / 'script.py'
    script.write_text("import _struct", encoding='utf-8')

    # Create a hook
    (tmp_path / 'hook-_struct.py').write_text("hiddenimports = ['ctypes']", encoding='utf-8')
    Analysis([str(script)], hookspath=[str(tmp_path)], excludes=['encodings', 'pydoc', 'xml', 'distutils'])


def test_issue_5131(monkeypatch, tmp_path):
    """
    While fixing the endless recursion when the package's __init__ module is an extension (see
    tests/unit/test_modulegraph_more.py::package_init_is_extension_*), another error occurred:
    PyInstaller.building._utils._load_code() tried to complete the source code for extension module - triggered by
    PYZ.assemble(), which is collecting all source files - caused by this being marked as "PYMODULE" in the TOC.
    """
    def get_imports(*args, **kwargs):
        # Our faked binary does not match the expected file-format for all platforms, thus the resp. code might crash.
        # Simply ignore this.
        try:
            return orig_get_imports(*args, **kwargs)
        except Exception:
            return []

    orig_get_imports = bindepend.get_imports
    monkeypatch.setattr(bindepend, "get_imports", get_imports)

    # On macOS, we need to similarly override `osxutils.get_macos_sdk_version`.
    if compat.is_darwin:
        from PyInstaller.utils import osx as osxutils

        def get_macos_sdk_version(*args, **kwargs):
            try:
                return orig_get_macos_sdk_version(*args, **kwargs)
            except Exception:
                return (10, 9, 0)  # Minimum version expected by check in Analysis.

        orig_get_macos_sdk_version = osxutils.get_macos_sdk_version
        monkeypatch.setattr(osxutils, "get_macos_sdk_version", get_macos_sdk_version)

    # Set up fake CONF for Analysis
    monkeypatch.setattr(
        'PyInstaller.config.CONF', {
            'workpath': str(tmp_path),
            'spec': str(tmp_path),
            'warnfile': str(tmp_path / 'warn.txt'),
            'dot-file': str(tmp_path / 'imports.dot'),
            'xref-file': str(tmp_path / 'imports.xref'),
            'hiddenimports': [],
            'specnm': 'issue_5131_script',
            'code_cache': dict(),
        }
    )

    # Speedup: avoid analyzing base_library.zip
    monkeypatch.setattr(analysis, 'PY3_BASE_MODULES', [])

    pkg = tmp_path / 'mypkg'
    pkg.mkdir()

    init = pkg / f"__init__{importlib.machinery.EXTENSION_SUFFIXES[0]}"
    init.write_bytes(b'\0\0\0\0\0\0\0\0\0\0\0\0' * 20)

    script = tmp_path / 'script.py'
    script.write_text("import mypkg", encoding='utf-8')

    a = Analysis([str(script)], excludes=['encodings', 'pydoc', 'xml', 'distutils'])
    PYZ(a.pure)


def test_issue_4141(pyi_builder):
    extra_path = _MODULES_DIR / 'pyi_issue_4141'
    pyi_builder.test_script(
        'pyi_issue_4141.py', app_name="main", run_from_path=True, pyi_args=['--path', str(extra_path)]
    )


def test_5734():
    """
    In a regression this will raise a:
        FileNotFoundError: [Errno 2] No such file or directory: b'liblibc.a'
    on some Linux/gcc combinations.
    """
    from PyInstaller.depend.utils import _resolveCtypesImports
    _resolveCtypesImports(["libc"])


def test_5797(pyi_builder):
    """
    Ensure that user overriding os and/or sys (using them as variables) in the global namespace does not affect
    PyInstaller's ctype hooks, installed during bootstrap, which is scenario of the issue #5797.
    """
    pyi_builder.test_source(
        """
        import ctypes

        # Overwrite sys and os in global namespace
        sys = 'watever'
        os = 'something'

        # Use a ctypes.CDLL, which in turn calls _frozen_name() from PyInstaller's ctypes bootstrap hooks.
        try:
            ctypes.CDLL('nonexistent')
        except Exception as e:
            # We expect PyInstallerImportError to be raised
            exception_name = type(e).__name__
            assert exception_name == 'PyInstallerImportError', f"Unexpected exception type: {exception_name}"
        """
    )



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# This test code is taken from the example code for the `future` library, with a few modifications to allow execution on
# 32-bit platforms: http://python-future.org/overview.html#code-examples

from builtins import (bytes, str, super, range, zip, round, int, pow, object)

# Backported Py3 bytes object
b = bytes(b'ABCD')
assert list(b) == [65, 66, 67, 68]
assert repr(b) == "b'ABCD'"
# These raise TypeErrors:
try:
    b + 'EFGH'
except TypeError:
    pass
else:
    assert False, "`bytes + str` did not raise TypeError"

try:
    bytes(b',').join(['Fred', 'Bill'])
except TypeError:
    pass
else:
    assert False, "`bytes.join([str, str])` did not raise TypeError"

# Backported Py3 str object
s = str('ABCD')
assert s != bytes(b'ABCD')
assert isinstance(s.encode('utf-8'), bytes)
assert isinstance(b.decode('utf-8'), str)
assert repr(s) == "'ABCD'"  # consistent repr with Py3 (no u prefix)
# These raise TypeErrors:
try:
    bytes(b'B') in s
except TypeError:
    pass
else:
    assert False, "`bytes in str` did not raise TypeError"

try:
    s.find(bytes(b'A'))
except TypeError:
    pass
else:
    assert False, "`str.find(bytes)` did not raise TypeError"


# New zero-argument super() function:
class VerboseList(list):
    def append(self, item):
        print('Adding an item')
        super().append(item)


# Fix: this fails on 32-bit Python. The traceback::
#
#        E:\pyinstaller>python tests\functional\scripts\pyi_future.py
#     Traceback (most recent call last):
#      File "tests\functional\scripts\pyi_future.py", line 66, in <module>
#        for i in range(10**15)[:10]:
#      File "C:\Users\user\python-2.7.10\lib\site-packages\future\types\newrange.py", line 122, in __getitem__
#        return self.__getitem_slice(index)
#      File "C:\Users\user\python-2.7.10\lib\site-packages\future\types\newrange.py", line 134, in __getitem_slice
#        scaled_indices = (self._step * n for n in slce.indices(self._len))
#     OverflowError: cannot fit 'long' into an index-sized integer
#
# So, pick a smaller (32-bit capable) range to iterate over.
#
# New iterable range object with slicing support
for i in range(2**30)[:10]:
    pass

# Other iterators: map, zip, filter
my_iter = zip(range(3), ['a', 'b', 'c'])
assert my_iter != list(my_iter)

# The round() function behaves as it does in Python 3, using "Banker's Rounding" to the nearest even last digit:
assert round(0.1250, 2) == 0.12

# pow() supports fractional exponents of negative numbers like in Py3:
z = pow(-1, 0.5)

# Compatible output from isinstance() across Py2/3:
assert isinstance(2**64, int)  # long integers
assert isinstance('blah', str)
assert isinstance('blah', str)  # only if unicode_literals is in effect


# Py3-style iterators written as new-style classes (subclasses of future.types.newobject) are automatically
# backward-compatible with Py2:
class Upper(object):
    def __init__(self, iterable):
        self._iter = iter(iterable)

    def __next__(self):  # note the Py3 interface
        return next(self._iter).upper()

    def __iter__(self):
        return self


assert list(Upper('hello')) == list('HELLO')



# -----------------------------------------------------------------------------
# Copyright (c) 2019-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
# -----------------------------------------------------------------------------

# This script establishes a simple communication with the bootloader to test the functions.
# If an error is detected, the program is closed before the timeout and thus counts as failed.
# The error can occur in different areas.

from time import sleep


def main():
    # Init pyi_splash / connect to the bootloader.
    import pyi_splash

    # Simulate users program startup.
    sleep(1)
    pyi_splash.update_text("This is a test text")
    sleep(2)
    pyi_splash.update_text("Second time's a charm")

    # Close the splash screen to check if that works.
    sleep(1)
    pyi_splash.close()

    # Wait for the timeout.
    sleep(20)


if __name__ == '__main__':
    main()
